// To compile:
// csc /t:library CustomDispatch.cs

using System;
using System.Reflection;
using System.Globalization;
using System.Collections;
using System.Runtime.InteropServices;

public class CustomDispatch : IReflect
{
  private Type t;
  private Hashtable names;

  // Well-known HRESULT returned by IDispatch.Invoke:
  private const int DISP_E_MEMBERNOTFOUND = unchecked((int)0x80020003);

  // Constructor
  public CustomDispatch()
  {
    t = this.GetType();

    // Fill Hashtable with explicit DISPIDs and their member names
    names = new Hashtable();
    DispIdAttribute [] ids;
    foreach (MemberInfo mi in t.GetMembers())
    {
      ids = (DispIdAttribute []) mi.GetCustomAttributes(
        typeof(DispIdAttribute), false);

      if (ids.Length > 0) names.Add(ids[0].Value, mi.Name);
    }
  }

  // Called by CLR to get DISPIDs and names for properties
  PropertyInfo [] IReflect.GetProperties(BindingFlags bindingAttr)
  {
    return t.GetProperties(bindingAttr);
  }

  // Called by CLR to get DISPIDs and names for fields
  FieldInfo [] IReflect.GetFields(BindingFlags bindingAttr)  
  {
    return t.GetFields(bindingAttr);
  }

  // Called by CLR to get DISPIDs and names for methods
  MethodInfo [] IReflect.GetMethods(BindingFlags bindingAttr)  
  {
    return t.GetMethods(bindingAttr);
  }

  // Called by CLR to invoke a member
  object IReflect.InvokeMember(string name, BindingFlags invokeAttr, 
    Binder binder, object target, object [] args, 
    ParameterModifier [] modifiers, CultureInfo culture,
    string [] namedParameters)  
  {
    // Special case for any member marked with DispIdAttribute
    if (name.StartsWith("[DISPID="))
    {
      int dispid = Convert.ToInt32(name.Substring(8, name.Length-9), 10);
      name = (string)names[dispid];
    }
    // Special case for DISPID_EVALUATE (-5)
    else if (name.Equals("[DISPID=-5]"))
    {
      name = "Evaluate";
    }

    try
    {
      return t.InvokeMember(name, invokeAttr, binder, target,
        args, modifiers, culture, namedParameters);
    }
    catch (MissingMemberException ex)
    {
      throw new COMException(ex.Message, DISP_E_MEMBERNOTFOUND);
    }
  }

  FieldInfo IReflect.GetField(string name, BindingFlags bindingAttr)
  {
    return t.GetField(name, bindingAttr);
  }

  MemberInfo [] IReflect.GetMember(string name, BindingFlags bindingAttr)
  {
    return t.GetMember(name, bindingAttr);
  }

  MemberInfo [] IReflect.GetMembers(BindingFlags bindingAttr)  
  {
    return t.GetMembers(bindingAttr);
  }

  MethodInfo IReflect.GetMethod(string name, BindingFlags bindingAttr)  
  {
    return t.GetMethod(name, bindingAttr);
  }

  MethodInfo IReflect.GetMethod(string name, BindingFlags bindingAttr,
    Binder binder, Type [] types, ParameterModifier [] modifiers)  
  {
    return t.GetMethod(name, bindingAttr, binder, types, modifiers);
  }

  PropertyInfo IReflect.GetProperty(string name, BindingFlags bindingAttr,
    Binder binder, Type returnType, Type [] types, 
    ParameterModifier [] modifiers)
  {
    return t.GetProperty(name, bindingAttr, binder,
      returnType, types, modifiers);
  }

  PropertyInfo IReflect.GetProperty(string name, BindingFlags bindingAttr)
  {
    return t.GetProperty(name, bindingAttr);
  }

  Type IReflect.UnderlyingSystemType
  {
    get { return t.UnderlyingSystemType; }
  }
}