// To compile, first obtain an Interop Assembly for the Smart Tags type library.
// If a PIA doesn't exist, run TLBIMP on MSTAG.TLB to produce SmartTagLib.dll.
// Then, run:
// csc /t:library DevTag.cs /r:SmartTagLib.dll

using System;
using SmartTagLib;
using System.Runtime.InteropServices;
using Microsoft.Win32;
using System.Reflection;

namespace Sams.SmartTags
{
  // Enumeration of verbs
  public enum Verbs
  {
    LookupInterface = 1,
    LookupTechnology = 2
  }

  // Enumeration of smart tag types
  public enum TagTypes
  {
    Interface = 1,
    Technology = 2
  }

  // The recognizer class
  [ClassInterface(ClassInterfaceType.None)]
  public class DevTagRecognizer : ISmartTagRecognizer
  {
    [ComRegisterFunction]
    private static void RegisterFunction(Type t)
    {
      // Add a custom key to the registry
      Registry.CurrentUser.CreateSubKey(@"Software\Microsoft\Office\Common\Smart Tag\Recognizers\" + t.GUID.ToString("B")).Close();
    }
    
    [ComUnregisterFunction]
    private static void UnregisterFunction(Type t)
    {
      // Undo the work from the registration function
      Registry.CurrentUser.DeleteSubKey(@"Software\Microsoft\Office\Common\Smart Tag\Recognizers\" + t.GUID.ToString("B"), false);
    }

    // Return the ProgID for this class
    public string ProgId
    {
      get { return "Sams.SmartTags.DevTagRecognizer"; }
    }

    // Return a name that will appear in the Tools/Autocorrect Options/Smart Tags dialog
    public string get_Name(int LocaleID)
    {
      return "Developer Terminology Recognizer";
    }
    
    // Return an in-depth description
    public string get_Desc(int LocaleID)
    {
      return "Developer Terminology Recognizer recognizes programming terms";
    }

    // Return the number of smart tag types supported
    public int SmartTagCount
    {
      get { return Enum.GetValues(typeof(TagTypes)).Length; }
    }

    // Return the name of the smart tag type specified by SmartTagID
    public string get_SmartTagName(int SmartTagID)
    {
      if (SmartTagID == (int)TagTypes.Interface)
        return "schemas-sams-com/DevTag#interface";
      else if (SmartTagID == (int)TagTypes.Technology)
        return "schemas-sams-com/DevTag#technology";
      else
        throw new ArgumentOutOfRangeException("SmartTagID", SmartTagID, "The ID must equal TagTypes.Interface or TagTypes.Technology.");
    }

    // Return a URL where actions can be downloaded
    public string get_SmartTagDownloadURL(int SmartTagID)
    {
      // We don't have such a URL
      return "";
    }

    // Recognize the programmer terms
    public void Recognize(string Text, IF_TYPE DataType, int LocaleID, ISmartTagRecognizerSite RecognizerSite)
    {
      string [] words;
      int index;

      words = Text.Split(null);
      index = 1;

      foreach (string word in words)
      {
        if (word.StartsWith("I") && word.Length > 3)
          RecognizerSite.CommitSmartTag("schemas-sams-com/DevTag#interface", index, word.Length, null);
        else if (word == ".NET" || word == "COM")
          RecognizerSite.CommitSmartTag("schemas-sams-com/DevTag#technology", index, word.Length, null);

        index += word.Length + 1;
      }
    }
  }

  // The action class
  [ClassInterface(ClassInterfaceType.None)]
  public class DevTagAction : ISmartTagAction
  {
    [ComRegisterFunction]
    private static void RegisterFunction(Type t)
    {
      // Add a custom key to the registry
      Registry.CurrentUser.CreateSubKey(@"Software\Microsoft\Office\Common\Smart Tag\Actions\" + t.GUID.ToString("B")).Close();
    }
    
    [ComUnregisterFunction]
    private static void UnregisterFunction(Type t)
    {
      // Undo the work from the registration function
      Registry.CurrentUser.DeleteSubKey(@"Software\Microsoft\Office\Common\Smart Tag\Actions\" + t.GUID.ToString("B"), false);
    }

    // Return the ProgID for this class
    public string ProgId
    {
      get { return "Sams.SmartTags.DevTagAction"; }
    }

    // Return a simple name
    public string get_Name(int LocaleID)
    {
      return "Developer Terminology Actions";
    }
    
    // Return an in-depth description
    public string get_Desc(int LocaleID)
    {
      return "Provides actions for programming terms";
    }

    // Return the number of smart tag types supported
    public int SmartTagCount
    {
      get { return Enum.GetValues(typeof(TagTypes)).Length; }
    }

    // Return the name of the smart tag type specified by SmartTagID
    public string get_SmartTagName(int SmartTagID)
    {
      if (SmartTagID == (int)TagTypes.Interface)
        return "schemas-sams-com/DevTag#interface";
      else if (SmartTagID == (int)TagTypes.Technology)
        return "schemas-sams-com/DevTag#technology";
      else
        throw new ArgumentOutOfRangeException("SmartTagID", SmartTagID, "The ID must equal TagTypes.Interface or TagTypes.Technology.");
    }

    // Return the caption of the smart tag type specified by SmartTagID
    public string get_SmartTagCaption(int SmartTagID, int LocaleID)
    {
      if (SmartTagID == (int)TagTypes.Interface)
        return "Developer Term: Interface";
      else if (SmartTagID == (int)TagTypes.Technology)
        return "Developer Term: Technology";
      else
        throw new ArgumentOutOfRangeException("SmartTagID", SmartTagID, "The ID must equal TagTypes.Interface or TagTypes.Technology.");
    }

    // Return the number of verbs supported for the smart tag type specified by SmartTagName
    public int get_VerbCount(string SmartTagName)
    {
      if (SmartTagName == "schemas-sams-com/DevTag#interface" ||
          SmartTagName == "schemas-sams-com/DevTag#technology")
        return 1;
      else
        throw new ArgumentException("Only tag names of 'interface' or 'technology' are supported.", SmartTagName);
    }

    // Return a unique identifier for the verb specified by VerbIndex and the tag specified by SmartTagName
    public int get_VerbID(string SmartTagName, int VerbIndex)
    {
      if (SmartTagName == "schemas-sams-com/DevTag#interface")
        return (int)Verbs.LookupInterface;
      else if (SmartTagName == "schemas-sams-com/DevTag#technology")
        return (int)Verbs.LookupTechnology;
      else
        throw new ArgumentException("Only tag names of 'interface' or 'technology' are supported.", SmartTagName);
    }

    // Return a caption for the verb specified by VerbID
    public string get_VerbCaptionFromID(int VerbID, string ApplicationName, int LocaleID)
    {
      if (VerbID == (int)Verbs.LookupInterface)
        return "Find online documentation for this interface";
      else if (VerbID == (int)Verbs.LookupTechnology)
        return "Find online documentation for this technology";
      else
        throw new ArgumentOutOfRangeException("VerbID", VerbID, "The ID must equal Verbs.LookupInterface or Verbs.LookupTechnology.");
    }

    // Return a name for the verb specified by VerbID
    public string get_VerbNameFromID(int VerbID)
    {
      return Enum.Format(typeof(Verbs), VerbID, "G");
    }

    // Act upon a verb
    public void InvokeVerb(int VerbID, string ApplicationName, object Target, ISmartTagProperties Properties, string Text, string XML)
    {
      object explorer;

      if (VerbID == (int)Verbs.LookupInterface || VerbID == (int)Verbs.LookupTechnology)
      {
        explorer = Activator.CreateInstance(Type.GetTypeFromProgID("InternetExplorer.Application"));
        explorer.GetType().InvokeMember("Navigate2", BindingFlags.InvokeMethod, null, explorer, new object[]{"http://search.microsoft.com/us/dev/default.asp?qu=" + Text + 
        "&boolean=ALL&nq=NEW&so=RECCNT&ig=01&i=00&i=01&i=02&i=03&i=04&i=05&i=06&i=07&i=08&i=09&i=10&i=11&i=12&i=13&i=14&i=15&i=16&i=17&i=18&ig=02&i=19&i=20&i=21&i=22&i=23&i=24&i=25&i=26&i=27&i=28&i=29&i=30&i=31&i=32&i=33&i=34&ig=03&i=35&i=36&i=37&i=38&i=39&i=40&i=41&ig=04&i=42&i=43&i=44&i=45&i=46&i=47&i=48&i=49&i=50&i=51&i=52&i=53&i=54&i=55&i=56&i=57&i=58&ig=05&ig=06&ig=07&p=1&nq=NEW",
        Type.Missing, Type.Missing, Type.Missing, Type.Missing});

        explorer.GetType().InvokeMember("Visible", BindingFlags.SetProperty, null, explorer, new object[]{true});
      }
      else
      {
        throw new ArgumentOutOfRangeException("VerbID", VerbID, "The ID must equal Verbs.LookupInterface or Verbs.LookupTechnology.");
      } 
    }
  }
}