' To compile, first obtain an Interop Assembly for the Smart Tags type library.
' If a PIA doesn't exist, run TLBIMP on MSTAG.TLB to produce SmartTagLib.dll.
' Then, run:
' vbc /t:library DevTag.vb /r:SmartTagLib.dll

Imports System
Imports SmartTagLib
Imports Microsoft.VisualBasic
Imports System.Runtime.InteropServices
Imports Microsoft.Win32

Namespace Sams.SmartTags

  ' Enumeration of verbs
  Public Enum Verbs
    LookupInterface = 1
    LookupTechnology = 2
  End Enum

  ' Enumeration of smart tag types
  Public Enum TagTypes
    IsInterface = 1
    IsTechnology = 2
  End Enum

  ' The recognizer class
  <ClassInterface(ClassInterfaceType.None)> Public Class DevTagRecognizer
    Implements ISmartTagRecognizer

    ' Add a custom key to the registry
    <ComRegisterFunction()> _
    Private Shared Sub RegisterFunction(ByVal t As Type)
      Registry.CurrentUser.CreateSubKey( _
        "Software\Microsoft\Office\Common\Smart Tag\Recognizers\" _
        & t.GUID.ToString("B")).Close()
    End Sub

    ' Undo the work from the registration function
    <ComUnregisterFunction()> _
    Private Shared Sub UnregisterFunction(ByVal t As Type)
      Registry.CurrentUser.DeleteSubKey( _
        "Software\Microsoft\Office\Common\Smart Tag\Recognizers\" _
        & t.GUID.ToString("B"), False)
    End Sub

    ' Return the ProgID for this class
    Public ReadOnly Property ProgId() As String _
      Implements ISmartTagRecognizer.ProgId
      Get
        ProgId = "Sams.SmartTags.DevTagRecognizer"
      End Get
    End Property

    ' Return a name to be displayed in the 
    ' Tools/Autocorrect Options/Smart Tags dialog
    Public ReadOnly Property Name(ByVal LocaleID As Integer) As String _
      Implements ISmartTagRecognizer.Name
      Get
        Name = "Developer Terminology Recognizer"
      End Get
    End Property

    ' Return a description
    Public ReadOnly Property Desc(ByVal LocaleID As Integer) As String _
      Implements ISmartTagRecognizer.Desc
      Get
        Desc = "The Developer Terminology Recognizer recognizes two " & _
          "kinds of programming terms - interfaces and technologies."
      End Get
    End Property

    ' Return the number of smart tag types supported
    Public ReadOnly Property SmartTagCount() As Integer _
      Implements ISmartTagRecognizer.SmartTagCount
      Get
        SmartTagCount = [Enum].GetValues(GetType(TagTypes)).Length
      End Get
    End Property

    ' Return the name of the smart tag type specified by SmartTagID
    Public ReadOnly Property SmartTagName(ByVal SmartTagID As Integer) _
      As String Implements ISmartTagRecognizer.SmartTagName
      Get
        If (SmartTagID = TagTypes.IsInterface) Then
          SmartTagName = "schemas-sams-com/DevTag#interface"
        ElseIf (SmartTagID = TagTypes.IsTechnology) Then
          SmartTagName = "schemas-sams-com/DevTag#technology"
        Else
          Throw New ArgumentOutOfRangeException("SmartTagID", _
            SmartTagID, _
            "Expected TagTypes.IsInterface or TagTypes.IsTechnology.")
        End If
      End Get
    End Property

    ' Return a URL where actions can be downloaded
    Public ReadOnly Property SmartTagDownloadURL( _
      ByVal SmartTagID As Integer) As String _
      Implements ISmartTagRecognizer.SmartTagDownloadURL
      Get
        ' We don't have such a URL
        SmartTagDownloadURL = ""
      End Get
    End Property

    ' Recognize the programmer terms
    Public Sub Recognize(ByVal Text As String, _
      ByVal DataType As IF_TYPE, ByVal LocaleID As Integer, _
      ByVal RecognizerSite As ISmartTagRecognizerSite) _
      Implements ISmartTagRecognizer.Recognize

      Dim word As String 
      Dim words() As String = Text.Split()
      Dim index As Integer = 1

      For Each word In words
        If (Char.IsPunctuation(word.Chars(word.Length-1))) Then 
          word = word.Substring(0, word.Length-1)
        End If

        If (word.StartsWith("I") And word.Length > 3) Then
          RecognizerSite.CommitSmartTag("schemas-sams-com/" & _
            "DevTag#interface", index, word.Length, Nothing)
        ElseIf (word = ".NET" Or word = "COM") Then
          RecognizerSite.CommitSmartTag("schemas-sams-com/" & _
            "DevTag#technology", index, word.Length, Nothing)
        End If
        index = index + word.Length + 1
      Next word
    End Sub
  End Class

  ' The action class
  <ClassInterface(ClassInterfaceType.None)> Public Class DevTagAction
    Implements ISmartTagAction

    ' Add a custom key to the registry
    <ComRegisterFunction()> _
    Private Shared Sub RegisterFunction(ByVal t As Type)
      Registry.CurrentUser.CreateSubKey( _
        "Software\Microsoft\Office\Common\Smart Tag\Actions\" _
        & t.GUID.ToString("B")).Close()
    End Sub

    ' Undo the work from the registration function
    <ComUnregisterFunction()> _
    Private Shared Sub UnregisterFunction(ByVal t As Type)
      Registry.CurrentUser.DeleteSubKey( _
      "Software\Microsoft\Office\Common\Smart Tag\Actions\" _
      & t.GUID.ToString("B"), False)
    End Sub

    ' Return the ProgID for this class
    Public ReadOnly Property ProgId() As String _
      Implements ISmartTagAction.ProgId
      Get
        ProgId = "Sams.SmartTags.DevTagAction"
      End Get
    End Property

    ' Return a simple name
    Public ReadOnly Property Name(ByVal LocaleID As Integer) As String _
      Implements ISmartTagAction.Name
      Get
        Name = "Developer Terminology Actions"
      End Get
    End Property

    ' Return a description
    Public ReadOnly Property Desc(ByVal LocaleID As Integer) As String _
      Implements ISmartTagAction.Desc
      Get
        Desc = "Provides actions for programming terms"
      End Get
    End Property

    ' Return the number of smart tag types supported
    Public ReadOnly Property SmartTagCount() As Integer _
      Implements ISmartTagAction.SmartTagCount
      Get
        SmartTagCount = [Enum].GetValues(GetType(TagTypes)).Length
      End Get
    End Property

    ' Return the name of the smart tag type specified by SmartTagID
    Public ReadOnly Property SmartTagName(ByVal SmartTagID As Integer) _
      As String Implements ISmartTagAction.SmartTagName
      Get
        If (SmartTagID = TagTypes.IsInterface) Then
          SmartTagName = "schemas-sams-com/DevTag#interface"
        ElseIf (SmartTagID = TagTypes.IsTechnology) Then
          SmartTagName = "schemas-sams-com/DevTag#technology"
        Else
          Throw New ArgumentOutOfRangeException("SmartTagID", _
            SmartTagID, _
            "Expected TagTypes.IsInterface or TagTypes.IsTechnology.")
        End If
      End Get
    End Property

    ' Return the caption of the smart tag type specified by SmartTagID
    Public ReadOnly Property SmartTagCaption( _
      ByVal SmartTagID As Integer, ByVal LocaleID As Integer) As String _
      Implements ISmartTagAction.SmartTagCaption
      Get
        If (SmartTagID = TagTypes.IsInterface) Then
          SmartTagCaption = "Developer Term: Interface"
        ElseIf (SmartTagID = TagTypes.IsTechnology) Then
          SmartTagCaption = "Developer Term: Technology"
        Else
          Throw New ArgumentOutOfRangeException("SmartTagID", _
            SmartTagID, _
            "Expected TagTypes.IsInterface or TagTypes.IsTechnology.")
        End If
      End Get
    End Property

    ' Return the number of verbs supported for the 
    ' smart tag type specified by SmartTagName
    Public ReadOnly Property VerbCount(ByVal SmartTagName As String) _
      As Integer Implements ISmartTagAction.VerbCount
      Get
        If (SmartTagName = "schemas-sams-com/DevTag#interface" Or _
          SmartTagName = "schemas-sams-com/DevTag#technology") Then
          VerbCount = 1
        Else
          Throw New ArgumentException("Only tag names of 'interface'" & _
          " or 'technology' are supported.", SmartTagName)
        End If
      End Get
    End Property

    ' Return a unique identifier for the verb 
    ' specified by VerbIndex and the tag specified by SmartTagName
    Public ReadOnly Property VerbID(ByVal SmartTagName As String, _
      ByVal VerbIndex As Integer) As Integer _
      Implements ISmartTagAction.VerbID
      Get
        If (SmartTagName = "schemas-sams-com/DevTag#interface") Then
          VerbID = Verbs.LookupInterface
        ElseIf (SmartTagName = "schemas-sams-com/DevTag#technology") Then
          VerbID = Verbs.LookupTechnology
        Else
          Throw New ArgumentException("Only tag names of 'interface'" & _
            " or 'technology' are supported.", SmartTagName)
        End If
      End Get
    End Property

    ' Return a caption for the verb specified by VerbID
    Public ReadOnly Property VerbCaptionFromID( _
      ByVal VerbID As Integer, ByVal ApplicationName As String, _
      ByVal LocaleID As Integer) As String _
      Implements ISmartTagAction.VerbCaptionFromID
      Get
        If (VerbID = Verbs.LookupInterface) Then
          VerbCaptionFromID = _
            "Find online documentation for this interface"
        ElseIf (VerbID = Verbs.LookupTechnology) Then
          VerbCaptionFromID = _
            "Find online documentation for this technology"
        Else
          Throw New ArgumentOutOfRangeException("VerbID", VerbID, _
            "Expected Verbs.LookupInterface or Verbs.LookupTechnology.")
        End If
      End Get
    End Property

    ' Return a name for the verb specified by VerbID
    Public ReadOnly Property VerbNameFromID(ByVal VerbID As Integer) _
      As String Implements ISmartTagAction.VerbNameFromID
      Get
        VerbNameFromID = [Enum].Format(GetType(Verbs), VerbID, "G")
      End Get
    End Property

    ' Act upon a verb
    Public Sub InvokeVerb( _
      ByVal VerbID As Integer, ByVal ApplicationName As String, _
      ByVal Target As Object, ByVal Properties As ISmartTagProperties, _
      ByVal Text As String, ByVal XML As String) _
      Implements ISmartTagAction.InvokeVerb

      Dim explorer As Object

      If (VerbID = Verbs.LookupInterface Or _
        VerbID = Verbs.LookupTechnology) Then
        explorer = CreateObject("InternetExplorer.Application")
        explorer.Navigate2( _
          "http://search.microsoft.com/us/dev/default.asp?qu=" & Text & _
          "&boolean=ALL&nq=NEW&so=RECCNT&ig=01&i=00&i=01&i=02&i=03" & _
          "&i=04&i=05&i=06&i=07&i=08&i=09&i=10&i=11&i=12&i=13&i=14" & _
          "&i=15&i=16&i=17&i=18&ig=02&i=19&i=20&i=21&i=22&i=23&i=24" & _
          "&i=25&i=26&i=27&i=28&i=29&i=30&i=31&i=32&i=33&i=34&ig=03" & _
          "&i=35&i=36&i=37&i=38&i=39&i=40&i=41&ig=04&i=42&i=43&i=44" & _
          "&i=45&i=46&i=47&i=48&i=49&i=50&i=51&i=52&i=53&i=54&i=55" & _
          "&i=56&i=57&i=58&ig=05&ig=06&ig=07&p=1&nq=NEW")

        explorer.Visible = True
      Else
        Throw New ArgumentOutOfRangeException("VerbID", VerbID, _
          "Expected Verbs.LookupInterface or Verbs.LookupTechnology.")
      End If
    End Sub
  End Class
End Namespace