// To compile:
// csc /t:library Listing19_4cs.cs

using System;
using System.Runtime.InteropServices;

// Structs used by the PInvoke methods:
   
internal struct CONSOLE_SCREEN_BUFFER_INFO 
{
  internal COORD dwSize;
  internal COORD dwCursorPosition;
  internal ushort wAttributes;
  internal SMALL_RECT srWindow;
  internal COORD dwMaximumWindowSize;
}

internal struct COORD 
{
  internal short X;
  internal short Y;
}

internal struct SMALL_RECT 
{
  internal short Left;
  internal short Top;
  internal short Right;
  internal short Bottom;
}

// Constants used with PInvoke methods
internal class Constants
{
  // Control foreground color
  internal const ushort FOREGROUND_BLUE      = 0x01; 
  internal const ushort FOREGROUND_GREEN     = 0x02;
  internal const ushort FOREGROUND_RED       = 0x04;
  internal const ushort FOREGROUND_INTENSITY = 0x08;

  // Control background color
  internal const ushort BACKGROUND_BLUE      = 0x10;
  internal const ushort BACKGROUND_GREEN     = 0x20;
  internal const ushort BACKGROUND_RED       = 0x40;
  internal const ushort BACKGROUND_INTENSITY = 0x80;

  // Standard input, output, and error
  internal const int STD_INPUT_HANDLE  = -10;
  internal const int STD_OUTPUT_HANDLE = -11;
  internal const int STD_ERROR_HANDLE  = -12;

  // Returned by GetStdHandle when an error occurs
  internal static readonly IntPtr INVALID_HANDLE_VALUE = new IntPtr(-1);
}

// The public ColorfulConsole class
public class ColorfulConsole 
{
  // Three PInvoke signatures:

  [DllImport("kernel32.dll")]
  private static extern IntPtr GetStdHandle(int nStdHandle);

  [DllImport("kernel32.dll")]
  private static extern bool GetConsoleScreenBufferInfo(
    IntPtr hConsoleOutput,
    out CONSOLE_SCREEN_BUFFER_INFO lpConsoleScreenBufferInfo);

  [DllImport("kernel32.dll")]
  private static extern bool SetConsoleTextAttribute(
    IntPtr hConsoleOutput, 
    ushort wAttributes);

  private static CONSOLE_SCREEN_BUFFER_INFO csbi = 
    new CONSOLE_SCREEN_BUFFER_INFO();

  // Two overloads of WriteLine:

  // Print text with the given foreground color and background color
  public static void WriteLine(string text, ForegroundColors foreColor,
    BackgroundColors backColor) 
  {
    IntPtr stdout = GetStdHandle(Constants.STD_OUTPUT_HANDLE);

    if (stdout == Constants.INVALID_HANDLE_VALUE)
      throw new Exception("Unable to get standard output handle");

    // Save existing console settings
    if (!GetConsoleScreenBufferInfo(stdout, out csbi))
      throw new Exception("Unable to get existing console settings");

    // Set the color
    if (!SetConsoleTextAttribute(stdout, 
      (ushort)((ushort)foreColor | (ushort)backColor)))
      throw new Exception("Unable to set console colors");

    // Write the text using System.Console
    Console.WriteLine(text);

    // Restore the old console settings
    if (!SetConsoleTextAttribute(stdout, csbi.wAttributes))
      throw new Exception("Unable to restore normal console colors");
  }

  // Print text with the given foreground color 
  // and current console background color
  public static void WriteLine(string text, ForegroundColors foreColor)
  {
    IntPtr stdout = GetStdHandle(Constants.STD_OUTPUT_HANDLE);

    if (stdout == Constants.INVALID_HANDLE_VALUE)
      throw new Exception("Unable to get standard output handle");

    // Retrieve the current background color
    if (!GetConsoleScreenBufferInfo(stdout, out csbi))
      throw new Exception("Unable to get existing console settings");

    BackgroundColors defaultBackground = 
      (BackgroundColors)(csbi.wAttributes & 0xF0);

    WriteLine(text, foreColor, defaultBackground);
  }
}

// Enumeration of foreground colors
public enum ForegroundColors : ushort
{
  Black = 0,
  Blue = Constants.FOREGROUND_BLUE,
  Green = Constants.FOREGROUND_GREEN,
  Cyan = Constants.FOREGROUND_BLUE | Constants.FOREGROUND_GREEN,
  Red = Constants.FOREGROUND_RED,
  Magenta = Constants.FOREGROUND_RED | Constants.FOREGROUND_BLUE,
  Brown = Constants.FOREGROUND_RED | Constants.FOREGROUND_GREEN,
  Gray = Constants.FOREGROUND_RED | Constants.FOREGROUND_BLUE | 
    Constants.FOREGROUND_GREEN,
  DarkGray = Constants.FOREGROUND_INTENSITY,
  LightBlue = Constants.FOREGROUND_BLUE | Constants.FOREGROUND_INTENSITY,
  LightGreen = Constants.FOREGROUND_GREEN | 
    Constants.FOREGROUND_INTENSITY,
  LightCyan = Constants.FOREGROUND_BLUE | Constants.FOREGROUND_GREEN | 
    Constants.FOREGROUND_INTENSITY,
  LightRed = Constants.FOREGROUND_RED | Constants.FOREGROUND_INTENSITY,
  LightMagenta = Constants.FOREGROUND_RED | Constants.FOREGROUND_BLUE |
    Constants.FOREGROUND_INTENSITY,
  Yellow = Constants.FOREGROUND_RED | Constants.FOREGROUND_GREEN | 
    Constants.FOREGROUND_INTENSITY,
  White = Constants.FOREGROUND_RED | Constants.FOREGROUND_BLUE |
    Constants.FOREGROUND_GREEN | Constants.FOREGROUND_INTENSITY
}

// Enumeration of background colors
public enum BackgroundColors : ushort
{
  Black = 0,
  Blue = Constants.BACKGROUND_BLUE,
  Green = Constants.BACKGROUND_GREEN,
  Cyan = Constants.BACKGROUND_BLUE | Constants.BACKGROUND_GREEN,
  Red = Constants.BACKGROUND_RED,
  Magenta = Constants.BACKGROUND_RED | Constants.BACKGROUND_BLUE,
  Brown = Constants.BACKGROUND_RED | Constants.BACKGROUND_GREEN,
  Gray = Constants.BACKGROUND_RED | Constants.BACKGROUND_BLUE |
    Constants.BACKGROUND_GREEN,
  DarkGray = Constants.BACKGROUND_INTENSITY,
  LightBlue = Constants.BACKGROUND_BLUE | Constants.BACKGROUND_INTENSITY,
  LightGreen = Constants.BACKGROUND_GREEN | 
    Constants.BACKGROUND_INTENSITY,
  LightCyan = Constants.BACKGROUND_BLUE | Constants.BACKGROUND_GREEN | 
    Constants.BACKGROUND_INTENSITY,
  LightRed = Constants.BACKGROUND_RED | Constants.BACKGROUND_INTENSITY,
  LightMagenta = Constants.BACKGROUND_RED | Constants.BACKGROUND_BLUE | 
    Constants.BACKGROUND_INTENSITY,
  Yellow = Constants.BACKGROUND_RED | Constants.BACKGROUND_GREEN | 
    Constants.BACKGROUND_INTENSITY,
  White = Constants.BACKGROUND_RED | Constants.BACKGROUND_BLUE | 
    Constants.BACKGROUND_GREEN | Constants.BACKGROUND_INTENSITY
}