// To compile:
// csc /unsafe /t:library /out:StreamMarshaler.dll StreamMarshaler.cs ComStream.cs ManagedIStream.cs

using System;
using System.IO;
using System.Runtime.InteropServices;

public class ComStream : Stream
{
  // The managed stream being wrapped
  UCOMIStream originalStream;

  // Constructor
  public ComStream(UCOMIStream stream)
  {
    if (stream != null)
    {
      originalStream = stream;
    }
    else
    {
      throw new ArgumentNullException("stream");
    }
  }

  // Finalizer
  ~ComStream()
  {
    Close();
  }

  // Property to get original stream object
  public UCOMIStream UnderlyingStream
  {
    get { return originalStream; }
  }

  // Reads a sequence of bytes from the current stream and advances the
  // current position within the stream by the number of bytes read
  public unsafe override int Read(byte [] buffer, int offset, int count)
  {
    if (originalStream == null)
      throw new ObjectDisposedException("originalStream");

    if (offset != 0)
      throw new NotSupportedException("Only a zero offset is supported.");

    int bytesRead;

    // Use unsafe code to easily get the address of a value type
    IntPtr address = new IntPtr(&bytesRead);

    originalStream.Read(buffer, count, address);

    return bytesRead;
  }

  // Writes a sequence of bytes to the current stream and advances the
  // current position within this stream by the number of bytes written
  public override void Write(byte [] buffer, int offset, int count)
  {
    if (originalStream == null)
      throw new ObjectDisposedException("originalStream");

    if (offset != 0)
      throw new NotSupportedException("Only a zero offset is supported.");

    // Pass "null" for the last parameter since we don't use the value
    originalStream.Write(buffer, count, IntPtr.Zero);
  }

  // Sets the position within the current stream
  public unsafe override long Seek(long offset, SeekOrigin origin)
  {
    if (originalStream == null)
      throw new ObjectDisposedException("originalStream");

    long position = 0;

    // Use unsafe code to easily get the address of a value type
    IntPtr address = new IntPtr(&position);

    // The enum values of SeekOrigin match the enum values of
    // STREAM_SEEK, so we can just cast the origin to an integer.
    originalStream.Seek(offset, (int)origin, address);

    return position;
  }

  // Returns the length, in bytes, of the stream
  public override long Length
  {
    get
    {
      if (originalStream == null)
        throw new ObjectDisposedException("originalStream");

      // Call IStream.Stat to retrieve info about the stream,
      // which includes the length. STATFLAG_NONAME means that we don't
      // care about the name (STATSTG.pwcsName), so there is no need for
      // the method to allocate memory for the string.

      STATSTG statstg;
      originalStream.Stat(out statstg, 1 /*STATFLAG_NONAME*/);
      return statstg.cbSize;
    }
  }

  // Determines the position within the current stream
  public override long Position
  {
    get { return Seek(0, SeekOrigin.Current); }
    set { Seek(value, SeekOrigin.Begin); }
  }

  // Sets the length of the current stream
  public override void SetLength(long value)
  {
    if (originalStream == null)
      throw new ObjectDisposedException("originalStream");
    
    originalStream.SetSize(value);
  }

  // Closes (disposes) the stream
  public override void Close()
  {
    if (originalStream != null)
    {
      originalStream.Commit(0); // STGC_DEFAULT
      Marshal.ReleaseComObject(originalStream);
      originalStream = null;
      GC.SuppressFinalize(this);
    }
  }

  // Updates the underlying data source or repository with the current
  // state of the buffer and then clears the buffer
  public override void Flush()
  {
    originalStream.Commit(0);
  }

  // Determines whether the current stream supports reading
  public override bool CanRead
  {
    get { return true; }
  }

  // Determines whether the current stream supports writing
  public override bool CanWrite
  {
    get { return true; }
  }

  // Determines whether the current stream supports seeking
  public override bool CanSeek
  {
    get { return true; }
  }
}