' To compile:
' vbc /t:library /out:StreamMarshaler.dll StreamMarshaler.vb ComStream.vb ManagedIStream.vb

Imports System
Imports System.IO
Imports System.Runtime.InteropServices

Public Class ComStream
  Inherits Stream

  ' The managed stream being wrapped
  Private originalStream As UCOMIStream

  ' Constructor
  Public Sub New(stream As UCOMIStream)
    If Not (stream Is Nothing) Then
      originalStream = stream
    Else
      Throw New ArgumentNullException("stream")
    End If
  End Sub
   
  ' Finalizer
  Protected Overloads Overrides Sub Finalize()
    Close()
    MyBase.Finalize()
  End Sub
   
  ' Property to get original stream object
   
  Public ReadOnly Property UnderlyingStream() As UCOMIStream
    Get
      Return originalStream
    End Get
  End Property

  ' Reads a sequence of bytes from the current stream and advances the
  ' current position within the stream by the number of bytes read
  Public Overrides Function Read(buffer() As Byte, offset As Integer, count As Integer) As Integer
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 

    If offset <> 0 Then
      Throw New NotSupportedException("Only a zero offset is supported.")
    End If 

    ' Use GCHandle to take the address of a value type (unlike the C# version)
    Dim bytesRead As Integer
    Dim box As Object = bytesRead
    Dim handle As GCHandle
    Try
      handle = GCHandle.Alloc(box, GCHandleType.Pinned)
      Dim address As IntPtr = handle.AddrOfPinnedObject()

      originalStream.Read(buffer, count, address)
    Finally
      If (handle.IsAllocated) Then handle.Free()
    End Try      
    Read = CInt(box)
  End Function

  ' Writes a sequence of bytes to the current stream and advances the
  ' current position within this stream by the number of bytes written
  Public Overrides Sub Write(buffer() As Byte, offset As Integer, count As Integer)
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 

    If offset <> 0 Then
      Throw New NotSupportedException("Only a zero offset is supported.")
    End If 

    ' Pass "null" for the last parameter since we don't use the value
    originalStream.Write(buffer, count, IntPtr.Zero)
  End Sub

  ' Sets the position within the current stream
  Public Overrides Function Seek(offset As Long, origin As SeekOrigin) As Long
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 

    ' Use GCHandle to take the address of a value type (unlike the C# version)
    Dim position As Long = 0
    Dim box As Object = position
    Dim handle As GCHandle
    Try
      handle = GCHandle.Alloc(box, GCHandleType.Pinned)
      Dim address As IntPtr = handle.AddrOfPinnedObject()

      ' The enum values of SeekOrigin match the enum values of
      ' STREAM_SEEK, so we can just cast the origin to an integer.
      originalStream.Seek(offset, CInt(origin), address)
    Finally
      If (handle.IsAllocated) Then handle.Free()
    End Try      
    Seek = CType(box, Long)
  End Function

  ' Returns the length, in bytes, of the stream
  Public Overrides ReadOnly Property Length() As Long
    Get
      If originalStream Is Nothing Then
        Throw New ObjectDisposedException("originalStream")
      End If 

      ' Call IStream.Stat to retrieve info about the stream,
      ' which includes the length. STATFLAG_NONAME means that we don't
      ' care about the name (STATSTG.pwcsName), so there is no need for
      ' the method to allocate memory for the string.
      Dim statstg As STATSTG
      originalStream.Stat(statstg, 1) 'STATFLAG_NONAME
      Return statstg.cbSize
    End Get
  End Property

  ' Determines the position within the current stream
  Public Overrides Property Position() As Long
    Get
      Return Seek(0, SeekOrigin.Current)
    End Get
    Set
      Seek(value, SeekOrigin.Begin)
    End Set
  End Property

  ' Sets the length of the current stream
  Public Overrides Sub SetLength(value As Long)
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 
    originalStream.SetSize(value)
  End Sub

  ' Closes (disposes) the stream
  Public Overrides Sub Close()
    If Not (originalStream Is Nothing) Then
      originalStream.Commit(0) ' STGC_DEFAULT
      Marshal.ReleaseComObject(originalStream)
      originalStream = Nothing
      GC.SuppressFinalize(Me)
    End If
  End Sub

  ' Updates the underlying data source or repository with the current
  ' state of the buffer and then clears the buffer
  Public Overrides Sub Flush()
    originalStream.Commit(0)
  End Sub

  ' Determines whether the current stream supports reading
  Public Overrides ReadOnly Property CanRead() As Boolean
    Get
      Return True
    End Get
  End Property 

  ' Determines whether the current stream supports writing
  Public Overrides ReadOnly Property CanWrite() As Boolean
    Get
      Return True
    End Get
  End Property 

  ' Determines whether the current stream supports seeking
  Public Overrides ReadOnly Property CanSeek() As Boolean
    Get
      Return True
    End Get
  End Property

End Class