' To compile:
' vbc /t:library /out:StreamMarshaler.dll StreamMarshaler.vb ComStream.vb ManagedIStream.vb

Imports System
Imports System.IO
Imports System.Runtime.InteropServices

<ClassInterface(ClassInterfaceType.None)> _
Public Class ManagedIStream
  Implements UCOMIStream
  Implements IDisposable

  ' The managed stream being wrapped
  Private originalStream As Stream
   
  ' Constructor
  Public Sub New(stream As Stream)
    If Not (stream Is Nothing) Then
      originalStream = stream
    Else
      Throw New ArgumentNullException("stream")
    End If
  End Sub
   
  ' Finalizer
  Protected Overloads Overrides Sub Finalize()
    Close()
    MyBase.Finalize()
  End Sub
 
  ' Property to get original stream object
  Public ReadOnly Property UnderlyingStream() As Stream
    Get
      If originalStream Is Nothing Then
        Throw New ObjectDisposedException("originalStream")
      End If 
      Return originalStream
    End Get
  End Property

  ' Reads a specified number of bytes from the stream object 
  ' into memory starting at the current seek pointer
  Public Sub Read(pv() As Byte, cb As Integer, pcbRead As IntPtr) Implements UCOMIStream.Read
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 
    If pcbRead.Equals(IntPtr.Zero) Then
      ' User isn't interested in how many bytes were read
      originalStream.Read(pv, 0, cb)
    Else
      Marshal.WriteInt32(pcbRead, originalStream.Read(pv, 0, cb))
    End If
  End Sub

  ' Writes a specified number of bytes into the stream object 
  ' starting at the current seek pointer
  Public Sub Write(pv() As Byte, cb As Integer, pcbWritten As IntPtr) Implements UCOMIStream.Write
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 
    If pcbWritten.Equals(IntPtr.Zero) Then
      ' User isn't interested in how many bytes were written
      originalStream.Write(pv, 0, cb)
    Else
      Dim originalPosition As Long = originalStream.Position
      originalStream.Write(pv, 0, cb)
      Marshal.WriteInt32(pcbWritten, CInt(originalStream.Position - originalPosition))
    End If
  End Sub

  ' Changes the seek pointer to a new location relative to the beginning
  ' of the stream, the end of the stream, or the current seek pointer
  Public Sub Seek(dlibMove As Long, dwOrigin As Integer, plibNewPosition As IntPtr) Implements UCOMIStream.Seek
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 
    ' The enum values of SeekOrigin match the enum values of
    ' STREAM_SEEK, so we can just cast the dwOrigin to a SeekOrigin
    If plibNewPosition.Equals(IntPtr.Zero) Then
      ' User isn't interested in new position
      originalStream.Seek(dlibMove, CType(dwOrigin, SeekOrigin))
    Else
      Marshal.WriteInt64(plibNewPosition, originalStream.Seek(dlibMove, CType(dwOrigin, SeekOrigin)))
    End If
  End Sub

  ' Changes the size of the stream object
  Public Sub SetSize(libNewSize As Long) Implements UCOMIStream.SetSize
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 
    originalStream.SetLength(libNewSize)
  End Sub

  ' Copies a specified number of bytes from the current seek pointer
  ' in the stream to the current seek pointer in another stream
  Public Sub CopyTo(pstm As UCOMIStream, cb As Long, pcbRead As IntPtr, pcbWritten As IntPtr) Implements UCOMIStream.CopyTo
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 
    Dim sourceBytes(cb) As Byte
    Dim currentBytesRead As Integer = 0
    Dim bytesWritten As Integer = 0

    While bytesWritten < cb
      currentBytesRead = originalStream.Read(sourceBytes, 0, CInt(cb - bytesWritten))

      ' Has the end of the stream been reached?
      If currentBytesRead = 0 Then
        Exit While
      End If 

      ' Stream.Write throws an exception if all bytes can't be written
      originalStream.Write(sourceBytes, 0, currentBytesRead)
      bytesWritten += currentBytesRead
    End While
      
    If Not pcbRead.Equals(IntPtr.Zero) Then
      Marshal.WriteInt64(pcbRead, bytesWritten)
    End If
    If Not pcbWritten.Equals(IntPtr.Zero) Then
      Marshal.WriteInt64(pcbWritten, bytesWritten)
    End If
  End Sub

  ' Ensures that any changes made to a stream object open in
  ' transacted mode are reflected in the parent storage object
  Public Sub Commit(grfCommitFlags As Integer) Implements UCOMIStream.Commit
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 
    originalStream.Flush()
  End Sub

  ' Discards all changes that have been made to a transacted 
  ' stream since the last IStream.Commit call
  Public Sub Revert() Implements UCOMIStream.Revert
    Throw New NotImplementedException("This stream does not support reverting.")
  End Sub

  ' Restricts access to a specified range of bytes in the stream
  Public Sub LockRegion(libOffset As Long, cb As Long, dwLockType As Integer) Implements UCOMIStream.LockRegion
    Throw New COMException("This stream does not support locking.", &H80030001)
  End Sub

  ' Removes the access restriction on a range of bytes
  Public Sub UnlockRegion(libOffset As Long, cb As Long, dwLockType As Integer) Implements UCOMIStream.UnlockRegion
    Throw New COMException("This stream does not support unlocking.", &H80030001)
  End Sub

  ' Retrieves the STATSTG structure for this stream
  Public Sub Stat(ByRef pstatstg As STATSTG, grfStatFlag As Integer) Implements UCOMIStream.Stat
    If originalStream Is Nothing Then
      Throw New ObjectDisposedException("originalStream")
    End If 
    pstatstg = New STATSTG()
    pstatstg.type = 2 ' STGTY_STREAM
    pstatstg.cbSize = originalStream.Length
    pstatstg.grfMode = 2 ' STGM_READWRITE;
    pstatstg.grfLocksSupported = 2 ' LOCK_EXCLUSIVE
  End Sub

  ' Creates a new stream object that references the same bytes as the 
  ' original stream but provides a separate seek pointer to those bytes
  Public Sub Clone(ByRef ppstm As UCOMIStream) Implements UCOMIStream.Clone
    Throw New NotImplementedException("This stream cannot be cloned.")
  End Sub

  ' Closes (disposes) the stream
  Public Sub Close()
    If Not (originalStream Is Nothing) Then
      originalStream.Close()
      originalStream = Nothing
      GC.SuppressFinalize(Me)
    End If
  End Sub

  ' IDisposable.Dispose
  Private Sub Dispose() Implements IDisposable.Dispose
    Close()
  End Sub

End Class