// To compile:
// cl /CLR StreamMarshaler.cpp

#using <mscorlib.dll> // Contains definition of Stream
#include "objidl.h"   // Contains definition of IStream
#include <vcclr.h>    // Contains the gcroot template

using namespace System;
using namespace System::IO;
using namespace System::Runtime::InteropServices;

// CustomCCW - A Custom COM-Callable Wrapper
class CustomCCW : public IStream
{
public:
  // Constructor
  CustomCCW(Stream *stream)
  {
    if (stream != NULL)
    {
      originalStream = stream;
      referenceCount = 0;

      // If the Stream is a FileStream, give a better mapping to IStream
      if (stream->GetType() == __typeof(FileStream))
        isFileStream = true;
      else
        isFileStream = false;
    }
    else
    {
      throw new ArgumentNullException(S"stream");
    }
  };

  // IUnknown.QueryInterface
  HRESULT STDMETHODCALLTYPE QueryInterface(const IID& iid, void** ppv)
  {
    if (iid == IID_IUnknown)
    {
      *ppv = static_cast<IStream*>(this);
    }
    else if (iid == IID_IStream)
    {
      *ppv = static_cast<IStream*>(this);
    }
    else
    {
      *ppv = NULL;
      return E_NOINTERFACE;
    }
    static_cast<IUnknown*>(*ppv)->AddRef();
    return S_OK;
  };

  // IUnknown.AddRef
  ULONG STDMETHODCALLTYPE AddRef()
  {
    InterlockedIncrement(&referenceCount);
    return referenceCount;
  };
  
  // IUnknown.Release
  ULONG STDMETHODCALLTYPE Release()
  {
    InterlockedDecrement(&referenceCount);
    if (referenceCount == 0)
    {
      Close();
      delete this;
      return 0;
    }
    return referenceCount;
  };

  // Method to get original stream object
  Object* GetUnderlyingObject()
  {
    return originalStream;
  };

  // Reads a specified number of bytes from the stream object 
  // into memory starting at the current seek pointer.
  HRESULT STDMETHODCALLTYPE Read(void *pv, ULONG cb, ULONG* pcbRead)
  {
    if (originalStream == NULL)
    {
      return Marshal::GetHRForException(
        new ObjectDisposedException(S"originalStream"));
    }
    try
    {
      // Allocate managed array to get the data from Stream.Read
      unsigned char array __gc[] = new unsigned char __gc[cb];

      if (pcbRead == NULL)
      {
         // User isn't interested in how many bytes were read
         *pcbRead = originalStream->Read(array, 0, cb);
      }
      else
      {
        originalStream->Read(array, 0, cb);
      }

      // Copy the contents of the managed array to the unmanaged buffer
      Marshal::Copy(array, 0, __nogc new IntPtr(pv), cb);
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }
    return S_OK;
  };

  // Writes a specified number of bytes into the stream object 
  // starting at the current seek pointer
  HRESULT STDMETHODCALLTYPE Write(const void *pv, ULONG cb, 
    ULONG* pcbWritten)
  {
    if (originalStream == NULL)
    {
      return Marshal::GetHRForException(
        new ObjectDisposedException(S"originalStream"));
    }
    try
    {
      // Allocate managed array to hold the data for Stream.Write
      unsigned char array __gc[] = new unsigned char __gc[cb];

      if (pcbWritten == NULL)
      {
        // User isn't interested in how many bytes were written
        originalStream->Write(array, 0, cb);
      }
      else
      {
        __int64 originalPosition = originalStream->Position;
        originalStream->Write(array, 0, cb);
        *pcbWritten = (ULONG)
          (originalStream->Position - originalPosition);
      }

      // Copy the contents of the unmanaged buffer to the managed array
      Marshal::Copy(__nogc new IntPtr((void*)pv), array, 0, cb);
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }
    return S_OK;
  };
              
  // Changes the seek pointer to a new location relative to the beginning
  // of the stream, the end of the stream, or the current seek pointer 
  HRESULT STDMETHODCALLTYPE Seek(LARGE_INTEGER dlibMove, DWORD dwOrigin, 
    ULARGE_INTEGER* plibNewPosition)
  {
    if (originalStream == NULL)
    {
      return Marshal::GetHRForException(
        new ObjectDisposedException(S"originalStream"));
    }
    try
    {
      // The enum values of SeekOrigin match the enum values of
      // STREAM_SEEK, so we can just cast the dwOrigin to a SeekOrigin

      if (plibNewPosition == NULL)
      {
        // User isn't interested in new position
        originalStream->Seek(dlibMove.QuadPart, (SeekOrigin)dwOrigin);
      }
      else
      {
        plibNewPosition->QuadPart = originalStream->Seek(
          dlibMove.QuadPart, (SeekOrigin)dwOrigin);
      }
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }
    return S_OK;
  };

  // Changes the size of the stream object. 
  HRESULT STDMETHODCALLTYPE SetSize(ULARGE_INTEGER libNewSize)
  {
    if (originalStream == NULL)
    {
      return Marshal::GetHRForException(new ObjectDisposedException(
        S"originalStream"));
    }
    try
    {
      originalStream->SetLength(libNewSize.QuadPart);
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }
    return S_OK;
  };

  // Copies a specified number of bytes from the current seek pointer
  // in the stream to the current seek pointer in another stream
  HRESULT STDMETHODCALLTYPE CopyTo(IStream* pstm, ULARGE_INTEGER cb, 
    ULARGE_INTEGER* pcbRead, ULARGE_INTEGER* pcbWritten)
  {
    if (originalStream == NULL)
    {
      return Marshal::GetHRForException(new ObjectDisposedException(
        S"originalStream"));
    }
    IntPtr sourceBytes;
    ULONG currentBytesRead = 0;
    int bytesWritten = 0;
    HRESULT hresult;

    try
    {
      // AllocHGlobal throws an OutOfMemoryException if it fails,
      // so we don't have to check the returned value
      sourceBytes = Marshal::AllocHGlobal((int)cb.QuadPart);

      while (bytesWritten < cb.QuadPart)
      {
        // Call CustomCCW's Read method
        hresult = Read(sourceBytes.ToPointer(), 
          cb.QuadPart - bytesWritten, &currentBytesRead);
        if (FAILED(hresult))
          return hresult;

        // Has the end of the stream been reached?
        if (currentBytesRead == 0) break;

        // Call CustomCCW's Write method
        hresult = Write(sourceBytes.ToPointer(), currentBytesRead, NULL);
        if (FAILED(hresult))
          return hresult;

        bytesWritten += currentBytesRead;
      }
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }
    __finally
    {
      Marshal::FreeHGlobal(sourceBytes);
    }

    if (pcbRead != NULL)   pcbRead->QuadPart = bytesWritten;
    if (pcbWritten != NULL) pcbWritten->QuadPart = bytesWritten;
    return S_OK;
  };

  // Ensures that any changes made to a stream object open in 
  // transacted mode are reflected in the parent storage object 
  HRESULT STDMETHODCALLTYPE Commit(DWORD grfCommitFlags)
  {
    if (originalStream == NULL)
    {
      return Marshal::GetHRForException(new ObjectDisposedException(
        S"originalStream"));
    }
    try
    {
      originalStream->Flush();
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }
    return S_OK;
  };

  // Discards all changes that have been made to a transacted 
  // stream since the last IStream::Commit call 
  HRESULT STDMETHODCALLTYPE Revert()
  {
    return E_NOTIMPL;
  };

  // Restricts access to a specified range of bytes in the stream
  HRESULT STDMETHODCALLTYPE LockRegion(ULARGE_INTEGER libOffset, 
    ULARGE_INTEGER cb, DWORD dwLockType)
  {
    if (originalStream == NULL)
    {
      return Marshal::GetHRForException(new ObjectDisposedException(
        S"originalStream"));
    }
    // We only provide an implementation if the Stream is a FileStream
    if (!isFileStream) return E_NOTIMPL;

    // We only support LOCK_EXCLUSIVE
    if (dwLockType != LOCK_EXCLUSIVE) return STG_E_INVALIDFUNCTION;

    try
    {
      // Call FileStream.Lock
      static_cast<FileStream*>((Stream*)originalStream)->
        Lock(libOffset.QuadPart, cb.QuadPart);
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }
    return S_OK;
  };

  // Removes the access restriction on a range of bytes
  HRESULT STDMETHODCALLTYPE UnlockRegion(ULARGE_INTEGER libOffset, 
    ULARGE_INTEGER cb, DWORD dwLockType)
  {
    if (originalStream == NULL)
    {
      return Marshal::GetHRForException(new ObjectDisposedException(
        S"originalStream"));
    }
    // We only provide an implementation if the Stream is a FileStream
    if (!isFileStream) return E_NOTIMPL;

    // We only support LOCK_EXCLUSIVE
    if (dwLockType != LOCK_EXCLUSIVE) return STG_E_INVALIDFUNCTION;

    try
    {
      // Call FileStream.Unlock
      static_cast<FileStream*>((Stream*)originalStream)->
        Unlock(libOffset.QuadPart, cb.QuadPart);
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }
    return S_OK;
  };

  // Retrieves the STATSTG structure for this stream.
  // If this is a FileStream, the mtime, ctime, atime, and pwcsName 
  // fields will contain information.
  HRESULT STDMETHODCALLTYPE Stat(tagSTATSTG* pstatstg, DWORD grfStatFlag)
  {
    if (originalStream == NULL)
    {
      return Marshal::GetHRForException(new ObjectDisposedException(
        S"originalStream"));
    }
    try
    {
      pstatstg->type = STGTY_STREAM;
      pstatstg->cbSize.QuadPart = originalStream->Length;
      pstatstg->grfMode = STGM_READWRITE;
      pstatstg->grfLocksSupported = LOCK_EXCLUSIVE;

      if (isFileStream)
      {
        _FILETIME creationTime;   
        _FILETIME lastAccessTime;
        _FILETIME lastWriteTime;
        IntPtr hFile = static_cast<FileStream*>((Stream*)originalStream)->
          Handle;
        BOOL b = GetFileTime((HANDLE)hFile.ToPointer(), &creationTime, 
          &lastAccessTime, &lastWriteTime);

        pstatstg->mtime = lastWriteTime;
        pstatstg->ctime = creationTime;
        pstatstg->atime = lastAccessTime;

        // Only give the name if the user passed STATFLAG_DEFAULT
        if (grfStatFlag == STATFLAG_DEFAULT)
        {
          pstatstg->pwcsName = (LPWSTR)Marshal::StringToCoTaskMemUni(
            static_cast<FileStream*>((Stream*)originalStream)->
            Name).ToPointer();
        }
      }
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }

    return S_OK;
  };

  // Creates a new stream object that references the same bytes as the 
  // original stream but provides a separate seek pointer to those bytes 
  HRESULT STDMETHODCALLTYPE Clone(IStream** ppstm)
  {
    return E_NOTIMPL;
  };

  // Closes (disposes) the stream
  HRESULT Close()
  {
    try
    {
      if (originalStream != NULL)
      {
        originalStream->Close();
        originalStream = NULL;
      }
    }
    catch (Exception* e)
    {
      return Marshal::GetHRForException(e);
    }
    return S_OK;
  }

private:
  gcroot<Stream*> originalStream;
  LONG referenceCount;
  bool isFileStream;
};

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------

// CustomRCW - A Custom Runtime-Callable Wrapper
public __gc class CustomRCW : public Stream, public ICustomAdapter
{
public:
  // Constructor
  CustomRCW(IStream* stream)
  {
    if (stream != NULL)
    {
      standardRCW = Marshal::GetObjectForIUnknown(stream);
    }
    else
    {
      throw new ArgumentNullException(S"stream");
    }
  };

  // Finalizer
  ~CustomRCW()
  {
    Close();
  };
 
  // Method to get original stream object
  Object* GetUnderlyingObject()
  {
    return standardRCW;
  };

  // Reads a sequence of bytes from the current stream and advances the
  // current position within the stream by the number of bytes read
  int Read(unsigned char buffer __gc[], int offset, int count)
  {
    if (standardRCW == NULL)
      throw new ObjectDisposedException(S"standardRCW");
    if (offset >= buffer->Length)
    {
      throw new ArgumentException(
        S"The offset must be less than the length of the buffer");
    }

    ULONG bytesRead;  // Number of bytes actually read
    GCHandle gcHandle;
    IStream* originalStream = NULL;
    
    try
    {
      // Pin the managed array
      gcHandle = GCHandle::Alloc(buffer, GCHandleType::Pinned);

      // Extract the interface pointer in the correct context
      originalStream = (IStream*)Marshal::GetComInterfaceForObject(
        standardRCW, __typeof(UCOMIStream)).ToPointer();

      // Call ISequentialStream.Read
      HRESULT hresult = originalStream->Read(
        (byte*)gcHandle.AddrOfPinnedObject().ToPointer() + offset, 
        count, &bytesRead);
      if (FAILED(hresult)) 
        Marshal::ThrowExceptionForHR(hresult);
    }
    __finally
    {
      if (gcHandle.IsAllocated) gcHandle.Free();
      if (originalStream != NULL) originalStream->Release();
    }

    return bytesRead;
  };

  // Writes a sequence of bytes to the current stream and advances the
  // current position within this stream by the number of bytes written
  void Write(unsigned char buffer __gc[], int offset, int count)
  {
    if (standardRCW == NULL)
      throw new ObjectDisposedException(S"standardRCW");

    if (offset >= buffer->Length)
    {
      throw new ArgumentException(
        S"The offset must be less than the length of the buffer");
    }

    ULONG cbWritten;   // Number of bytes actually written
    GCHandle gcHandle;
    IStream* originalStream = NULL; 

    try
    {
      // Pin the managed array
      GCHandle gcHandle = GCHandle::Alloc(buffer, GCHandleType::Pinned);

      // Extract the interface pointer in the correct context
      originalStream = (IStream*)Marshal::GetComInterfaceForObject(
        standardRCW, __typeof(UCOMIStream)).ToPointer();

      // Call ISequentialStream.Write
      HRESULT hresult = originalStream->Write(
        (byte*)gcHandle.AddrOfPinnedObject().ToPointer() + offset,
        count, &cbWritten);
      if (FAILED(hresult))
        Marshal::ThrowExceptionForHR(hresult);
    }
    __finally
    {
      if (gcHandle.IsAllocated) gcHandle.Free();
      if (originalStream != NULL) originalStream->Release();
    }
  };

  // Sets the position within the current stream
  __int64 Seek(__int64 offset, SeekOrigin origin)
  {
    if (standardRCW == NULL)
      throw new ObjectDisposedException(S"standardRCW");

    ULARGE_INTEGER plibNewPosition;  // Position after Seek
    LARGE_INTEGER off;
    off.QuadPart = offset;
    IStream* originalStream = NULL;

    try
    {
      // Extract the interface pointer in the correct context
      originalStream = (IStream*)Marshal::GetComInterfaceForObject(
        standardRCW, __typeof(UCOMIStream)).ToPointer();

      // The enum values of SeekOrigin match the enum values of
      // STREAM_SEEK, so we can just cast the origin to a DWORD
      HRESULT hresult = originalStream->Seek(off, (DWORD)origin, 
        &plibNewPosition);
      if (FAILED(hresult))
        Marshal::ThrowExceptionForHR(hresult);
    }
    __finally
    {
      if (originalStream != NULL) originalStream->Release();
    }

    return plibNewPosition.QuadPart;
  };

  // Returns the length, in bytes, of the stream
  __property __int64 get_Length() 
  {
    if (standardRCW == NULL)
      throw new ObjectDisposedException(S"standardRCW");

    tagSTATSTG statstg;
    IStream* originalStream = NULL;

    try
    {
      // Extract the interface pointer in the correct context
      originalStream = (IStream*)Marshal::GetComInterfaceForObject(
        standardRCW, __typeof(UCOMIStream)).ToPointer();

      // Call Stat to retrieve info about the stream, such as the length
      HRESULT hresult = originalStream->Stat(&statstg, STATFLAG_NONAME);
      if (FAILED(hresult))
        Marshal::ThrowExceptionForHR(hresult);
    }
    __finally
    {
      if (originalStream != NULL) originalStream->Release();
    }

    return statstg.cbSize.QuadPart;
  };

  // Determines the position within the current stream
  __property __int64 get_Position() 
  {
    return Seek(0, SeekOrigin::Current);
  };
  __property void set_Position(__int64 value) 
  {
    Seek(value, SeekOrigin::Begin);
  };

  // Sets the length of the current stream
  void SetLength(__int64 value)
  {
    if (standardRCW == NULL)
      throw new ObjectDisposedException(S"standardRCW");

    ULARGE_INTEGER val;
    val.QuadPart = value;
    IStream* originalStream = NULL;

    try
    {
      // Extract the interface pointer in the correct context
      originalStream = (IStream*)Marshal::GetComInterfaceForObject(
        standardRCW, __typeof(UCOMIStream)).ToPointer();

      HRESULT hresult = originalStream->SetSize(val);
      if (FAILED(hresult))
        Marshal::ThrowExceptionForHR(hresult);
    }
    __finally
    {
      if (originalStream != NULL) originalStream->Release();
    }
  };

  // Closes the current stream and releases any resources (such as
  // sockets and file handles) associated with the current stream
  void Close()
  {
    if (standardRCW != NULL)
    {
      IStream* originalStream = NULL;
      try
      {
        // Extract the interface pointer in the correct context
        originalStream = (IStream*)Marshal::GetComInterfaceForObject(
          standardRCW, __typeof(UCOMIStream)).ToPointer();
        originalStream->Commit(STGC_DEFAULT);
      }
      __finally
      {
        if (originalStream != NULL) originalStream->Release();
      }

      Marshal::ReleaseComObject(standardRCW);
      standardRCW = NULL;
      GC::SuppressFinalize(this);
    }
  };

  // Updates the underlying data source or repository with the current
  // state of the buffer and then clears the buffer
  void Flush()
  {
    IStream* originalStream = NULL;
    try
    {
      // Extract the interface pointer in the correct context
      originalStream = (IStream*)Marshal::GetComInterfaceForObject(
        standardRCW, __typeof(UCOMIStream)).ToPointer();
      originalStream->Commit(STGC_DEFAULT);
    }
    __finally
    {
      if (originalStream != NULL) originalStream->Release();
    }
  };

  // Determines whether the current stream supports reading
  __property bool get_CanRead() { return true; };

  // Determines whether the current stream supports seeking
  __property bool get_CanSeek() { return true; };

  // Determines whether the current stream supports writing
  __property bool get_CanWrite() { return true; };

private:
  Object* standardRCW;
};

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------

public __gc class StreamMarshaler : public ICustomMarshaler
{
private:
  static StreamMarshaler *marshaler = NULL;

public:
  Object* MarshalNativeToManaged(IntPtr pNativeData)
  {
    if (pNativeData == IntPtr::Zero) return NULL;
      return new CustomRCW(
        static_cast<IStream*>(pNativeData.ToPointer()));
  };

  IntPtr MarshalManagedToNative(Object* ManagedObj)
  {
    if (ManagedObj == NULL) return IntPtr::Zero;
    CustomCCW *ccw = new CustomCCW(static_cast<Stream*>(ManagedObj));
    IStream* intf = NULL;
    ccw->QueryInterface(IID_IStream, (void**)&intf);
    return (IntPtr)intf;
  }
 
  void CleanUpNativeData(IntPtr pNativeData)
  {
    Marshal::Release(pNativeData);
  };

  void CleanUpManagedData(Object* ManagedObj)
  {
    // Nothing to do
  };

  int GetNativeDataSize()
  {
    return -1;
  };

  static ICustomMarshaler* GetInstance(String* cookie)
  {
    // Always return the same instance
    if (marshaler != NULL) marshaler = new StreamMarshaler();
    return marshaler;
  };
};