// To compile:
// csc /out:Interop.exe Importer.cs Exporter.cs Registrar.cs ComPlusInstaller.cs MainForm.cs MiscTypes.cs /r:System.Windows.Forms.dll /r:System.dll /r:System.Drawing.dll /r:System.EnterpriseServices.dll

using System;
using System.IO;
using System.Reflection;
using System.Collections;
using System.Runtime.InteropServices;

// Import Options
public struct ImportOptions
{
  public TypeLibImporterFlags Flags;
  public bool UseKeyPairFromFile;
  public bool DelaySign;
  public string KeyFile;
  public string KeyContainer;
}

// Creates an assembly from a type library
public class Importer : ITypeLibImporterNotifySink
{
  private ImportOptions options;
  private TypeLibConverter converter;
  private IToolClient client;
  private Hashtable loadedAssemblies;
  private string outputDirectory;

  // Constructor
  public Importer(IToolClient client, Hashtable loadedAssemblies)
  {
    this.client = client;
    this.loadedAssemblies = loadedAssemblies;
    this.converter = new TypeLibConverter();
  }

  // Property to get/set options
  public ImportOptions Options
  {
    get { return options; }
    set { options = value; }
  }

  // Output directory where type libraries will be saved
  public string OutputDirectory
  {
    set { outputDirectory = value; }
  }

  // Creates an assembly with the given assembly name, namespace,
  // and version from the type library specified by typeLibFilename
  public Assembly Import(string typeLibFilename, string asmName, 
    string @namespace, Version version)
  {
    UCOMITypeLib typeLib;

    try
    {
      // Load the type library without registering it
      typeLib = Util.LoadTypeLibEx(typeLibFilename, REGKIND.REGKIND_NONE);
    }
    catch
    {
      throw new ApplicationException("Unable to load the type library '" +
        typeLibFilename + "'.");
    }

    return Import(typeLib, asmName, @namespace, version);
  }

  // Creates a type library with the given assembly name, namespace,
  // and version from the loaded input type library
  public Assembly Import(UCOMITypeLib typeLib, string asmName, 
    string @namespace, Version version)
  {
    string typeLibDisplayName;
    string piaName;
    string piaCodebase;
    Assembly returnAssembly;
    StrongNameKeyPair keyPair = null;
    byte [] keyData = null;
   
    // Before beginning import, check if a Primary Interop Assembly for
    // this type library is already registered
    if (GetPrimaryInteropAssembly(typeLib, out piaName, out piaCodebase))
    {
      if (client.AskYesNoQuestion("A Primary Interop Assembly is " +
        "already registered for the type library '" + 
        Util.GetDisplayNameForTypeLib(typeLib) + "' with full name '" +
        piaName + "' and codebase '" + piaCodebase + "'. Do you want " +
        "to load this assembly instead of importing a new one?", true))
      {
        returnAssembly = LoadPrimaryInteropAssembly(piaName, piaCodebase);

        client.SendMessage(MessageTypes.BeginFile, 
          "Primary Interop Assembly for " + 
          Util.GetDisplayNameForTypeLib(typeLib), null);
        client.SendMessage(MessageTypes.EndFile, 
          "Primary Interop Assembly for " + 
          Util.GetDisplayNameForTypeLib(typeLib), returnAssembly);

        loadedAssemblies[Marshal.GetTypeLibGuidForAssembly(
          returnAssembly).ToString() + 
          returnAssembly.GetName().Version.ToString(2)] = returnAssembly;
        return returnAssembly;
      }
    }

    // Construct a nice looking "display name" for the assembly
    if ((options.Flags & 
      TypeLibImporterFlags.PrimaryInteropAssembly) != 0)
    {
      typeLibDisplayName = "Primary ";
    }
    else
    {
      typeLibDisplayName = "";
    }

    typeLibDisplayName += "Interop Assembly for " +
      Util.GetDisplayNameForTypeLib(typeLib);

    // Choose a default assembly filename if none was chosen
    if (asmName == null)
      asmName = Marshal.GetTypeLibName(typeLib);
    
    asmName += ".dll";

    client.SendMessage(MessageTypes.BeginFile, typeLibDisplayName, 
      outputDirectory + Path.DirectorySeparatorChar + asmName);

    // Get the total number of types for the progress bar
    // by getting the number of types in the type library
    client.ProgressMaximum += typeLib.GetTypeInfoCount();

    // Extract key data if a key file is used
    if (options.KeyFile != null)
    {
      FileStream stream = null;
      try
      {
        stream = new FileStream(options.KeyFile, FileMode.Open,
          FileAccess.Read);
        keyData = new byte[stream.Length];
        stream.Read(keyData, 0, (int)stream.Length);
      }
      catch
      {
        throw new ApplicationException("Unable to read key data from '" +
          options.KeyFile + "'.");
      }
      finally
      {
        if (stream != null) stream.Close();
      }

      // Construct a SrongNameKeyPair if desired
      if (options.UseKeyPairFromFile)
      {
        keyPair = new StrongNameKeyPair(keyData);
        keyData = null;
      }
    }
    else if (options.KeyContainer != null)
    {
      // Construct a StrongNamePair from a key container
      keyPair = new StrongNameKeyPair(options.KeyContainer);
    }
    
    // If the user chose delay signing, extract only the public key
    // from the key pair
    if (options.DelaySign && keyPair != null)
    {
      keyData = keyPair.PublicKey;
      keyPair = null;
    }

    // Create an assembly from the type library
    returnAssembly = converter.ConvertTypeLibToAssembly(typeLib, asmName,
      options.Flags, this, keyData, keyPair, @namespace, version);

    client.SendMessage(MessageTypes.EndFile, typeLibDisplayName, 
      returnAssembly);
    loadedAssemblies[Marshal.GetTypeLibGuidForAssembly(
      returnAssembly).ToString() + 
      returnAssembly.GetName().Version.ToString(2)] = returnAssembly;
    return returnAssembly;
  }

  // For the input type library, find out if a Primary Interop Assembly
  // exists. If so, return its name and codebase via two out parameters
  private bool GetPrimaryInteropAssembly(UCOMITypeLib typeLib, 
    out string name, out string codebase)
  {
    TYPELIBATTR attributes;
    IntPtr attributesPtr = IntPtr.Zero;

    try
    {
      typeLib.GetLibAttr(out attributesPtr);
      attributes = (TYPELIBATTR)Marshal.PtrToStructure(attributesPtr,
        typeof(TYPELIBATTR));
    }
    finally
    {
      // As documented by ITypeLib, we must release the attributes
      if (attributesPtr != IntPtr.Zero)
        typeLib.ReleaseTLibAttr(attributesPtr);
    }
    
    // Now that we have all the type library information, we can
    // call TypeLibConverter.GetPrimaryInteropAssembly
    return converter.GetPrimaryInteropAssembly(attributes.guid,
      attributes.wMajorVerNum, attributes.wMinorVerNum, attributes.lcid,
      out name, out codebase);
  }

  // Return the version of a type library as a "Major.Minor" string
  private string GetTypeLibVersion(UCOMITypeLib typeLib)
  {
    TYPELIBATTR attributes;
    IntPtr attributesPtr = IntPtr.Zero;

    try
    {
      typeLib.GetLibAttr(out attributesPtr);
      attributes = (TYPELIBATTR)Marshal.PtrToStructure(attributesPtr,
        typeof(TYPELIBATTR));
    }
    finally
    {
      // As documented by ITypeLib, we must release the attributes
      if (attributesPtr != IntPtr.Zero)
        typeLib.ReleaseTLibAttr(attributesPtr);
    }
    
    // Now that we have all the type library information, we can
    // call TypeLibConverter.GetPrimaryInteropAssembly
    return attributes.wMajorVerNum + "." + attributes.wMinorVerNum;
  }

  // Load a Primary Interop Assembly from the given name and codebase.
  // The codebase may be null if none exists
  private Assembly LoadPrimaryInteropAssembly(string name,
    string codebase)
  {
    Assembly returnAssembly;

    try
    {
      // First attempt to load from its assembly name alone
      returnAssembly = Assembly.Load(name);
    }
    catch (FileNotFoundException)
    {
      // If the load failed, there better be a codebase so we can load
      // it from that location
      if (codebase != null)
        returnAssembly = Assembly.LoadFrom(codebase);
      else
        throw;
    }

    // Check that the loaded assembly is in fact a Primary
    // Interop Assembly
    if (returnAssembly.GetCustomAttributes(
      typeof(PrimaryInteropAssemblyAttribute), false).Length == 0)
    {
      throw new ApplicationException("The assembly '" + name + 
        "' is registered as a Primary Interop Assembly, but is not " +
        "marked with the PrimaryInteropAssembly custom attribute. " +
        "Please check your registration settings.");
    }

    return returnAssembly;
  }

  // Implementation of ITypeLibImporterNotifySink.ReportEvent
  public void ReportEvent(ImporterEventKind eventKind, int eventCode, 
    string eventMsg)
  {
    MessageTypes messageType;

    // Get the MessageType for the corresponding ImporterEventKind
    switch (eventKind)
    {
      case ImporterEventKind.ERROR_REFTOINVALIDTYPELIB:
        messageType = MessageTypes.InvalidRef;
        break;
      case ImporterEventKind.NOTIF_CONVERTWARNING:
        messageType = MessageTypes.Warning;
        break;
      case ImporterEventKind.NOTIF_TYPECONVERTED:
        messageType = MessageTypes.Notification;
        // Update the progress for type conversions
        client.Progress++;
        break;
      default:
        throw new ApplicationException(
          "Encountered unknown ImporterEventKind value.");
    }

    client.SendMessage(messageType, "[" + 
      DateTime.Now.ToString("hh:mm:ss.f") + " 0x" +
      eventCode.ToString("x") + "] " + eventMsg, null);
  }

  // Implementation of ITypeLibImporterNotifySink.ReportEvent
  public Assembly ResolveRef(object typeLib)
  {
    string piaName;
    string piaCodebase;
    UCOMITypeLib pTypeLib = (UCOMITypeLib)typeLib;

    // If we've already loaded the necessary assembly, return it
    Assembly returnAssembly = (Assembly)
      loadedAssemblies[Marshal.GetTypeLibGuid(pTypeLib).ToString() +
      GetTypeLibVersion(pTypeLib)];

    // If we're generating a Primary Interop Assembly and this dependent
    // assembly is not a Primary Interop Assembly, don't use it
    if (returnAssembly != null)
    {
      if ((options.Flags & 
        TypeLibImporterFlags.PrimaryInteropAssembly) != 0 &&
        returnAssembly.GetCustomAttributes(
        typeof(PrimaryInteropAssemblyAttribute), false).Length == 0)
      {
        // Erase the fact that we found an assembly
        returnAssembly = null;
      }
    }

    if (returnAssembly != null)
      return returnAssembly;
  
    // Before asking for a reference, attempt to find a Primary Interop 
    // Assembly for the referenced type library
    if (GetPrimaryInteropAssembly(pTypeLib, out piaName, out piaCodebase))
    {
      // We found one, so load it
      returnAssembly = LoadPrimaryInteropAssembly(piaName, piaCodebase);

      client.SendMessage(MessageTypes.Notification, "[" + 
        DateTime.Now.ToString("hh:mm:ss.f") + 
        "] Successfully referenced PIA for " + 
        Marshal.GetTypeLibName(pTypeLib), null);

      return returnAssembly;
    }

    // If we're creating a Primary Interop Assembly but no
    // Primary Interop Assembly exists for the referenced type library,
    // this is a fatal error
    if ((options.Flags & TypeLibImporterFlags.PrimaryInteropAssembly) 
      != 0)
      throw new ApplicationException("No registered Primary Interop " +
        "Assembly was found for the dependent type library '" +
        Marshal.GetTypeLibName(pTypeLib) + "' with LIBID " + 
        Marshal.GetTypeLibGuid(pTypeLib) + ". Either locate or create " +
        "one first, or choose not to create a Primary Interop Assembly " +
        "for the current type library.");

    // Ask the client for a filename to satisfy the reference
    string filename = client.GetTypeLibReference(
      Marshal.GetTypeLibGuid(pTypeLib), 
      Util.GetDisplayNameForTypeLib(pTypeLib));

    if (filename != null) 
    {
      // First treat the file as an assembly
      try
      {
        returnAssembly = Assembly.LoadFrom(filename);
        client.SendMessage(MessageTypes.BeginFile, 
          Util.GetDisplayNameForAssembly(returnAssembly), null);
        client.SendMessage(MessageTypes.EndFile, 
          Util.GetDisplayNameForAssembly(returnAssembly), returnAssembly);
        return returnAssembly;
      }
      catch
      {
           // Since LoadFrom didn't work, this must be a type library
      }
    }

    // Accept default assembly name, namespace, and version number      
    if (filename != null)
      return Import(filename, null, null, null);
    else
      return Import(pTypeLib, null, null, null);
  }
}