// To compile:
// csc /out:Interop.exe Importer.cs Exporter.cs Registrar.cs ComPlusInstaller.cs MainForm.cs MiscTypes.cs /r:System.Windows.Forms.dll /r:System.dll /r:System.Drawing.dll /r:System.EnterpriseServices.dll

using System;
using System.ComponentModel;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Collections;
using System.Reflection;
using System.Reflection.Emit;
using System.IO;
using System.EnterpriseServices;
using System.Threading;

// A TreeNode that contains a FilenameForSave property containing
// the filename with which a type library or assembly must be saved
public class CustomTreeNode : TreeNode
{
  private string filename;

  public CustomTreeNode(string text) : base(text) {}

  public string FilenameForSave
  {
    get { return filename; }
    set { filename = value; }
  }
}

// The class containing the application's entry point and the graphical
// user interface
public class MainForm : Form, IToolClient
{
  // General GUI items
  private TreeView treeView1;
  private TabControl tabControl1;
  private ProgressBar progressBar1;
  private OpenFileDialog openFileDialog1;
  private Container components = null;

  // Import tab items
  private TabPage importPage;
  private TextBox textVersion;
  private TextBox textNamespace;
  private TextBox textKeyContainer;
  private CheckBox checkPIA;
  private CheckBox checkArray;
  private CheckBox checkUnsafe;
  private CheckBox checkVersion;
  private CheckBox checkNamespace;
  private CheckBox checkDelaySign;
  private CheckBox checkSign;
  private RadioButton radioKeyPair;
  private RadioButton radioPublicKey;
  private RadioButton radioKeyContainer;
  private Button importOpenButton;
  private Button importSaveButton;
  private GroupBox groupSignAssembly;

  // Export tab items
  private TabPage exportPage;
  private TextBox textFilename;
  private TextBox textLibName;
  private TextBox textHelpFile;
  private TextBox textNames;
  private CheckBox checkFilename;
  private CheckBox checkLibName;
  private CheckBox checkHelpFile;
  private CheckBox checkNames;
  private Button exportOpenButton;
  private Button exportSaveButton;
  private System.Windows.Forms.Label exportNamesLabel;

  // Register tab items
  private TabPage registerPage;
  private CheckBox checkCodebase;
  private RadioButton radioUnregister;
  private RadioButton radioRegister;
  private Button registerNewButton;
  private Button registerSelectedButton;
  private GroupBox groupRegister;

  // COM+ tab items
  private TabPage comPlusPage;
  private TextBox textAppName;
  private TextBox textPartitionName;
  private TextBox textTypeLibName;
  private CheckBox checkAppName;
  private CheckBox checkPartitionName;
  private CheckBox checkTypeLibName;
  private CheckBox checkExistingTLB;
  private CheckBox checkNoReconfigure;
  private CheckBox checkComponentsOnly;
  private CheckBox checkRegisteredTLBs;
  private RadioButton radioFindOrCreate;
  private RadioButton radioCreate;
  private RadioButton radioUninstall;
  private RadioButton radioExistingApp;
  private Button comPlusRegisterButton;
  private GroupBox groupComPlusTarget;
  private GroupBox groupComPlusAction;

  // The 4 "tool classes"
  private Importer importer;
  private Exporter exporter;
  private Registrar registrar;
  private ComPlusInstaller comPlusInstaller;

  // Hashtables containing all assemblies and type libraries loaded
  private Hashtable loadedAssemblies;
  private Hashtable loadedTypeLibs;

  // Stack for managing treeView functionality
  private Stack filesBeingProcessed;
  
  // Background thread used for processing import and export
  private Thread backgroundThread;

  // Delegate used to send messages from the background thread
  private delegate void MessageDelegate(MessageTypes type, 
    string message, object data);

  // The entry point
  [STAThread]
  static void Main()
  {
    Application.Run(new MainForm());
  }

  // Constructor
  public MainForm()
  {
    InitializeComponent();

    filesBeingProcessed = new Stack();
    loadedAssemblies = new Hashtable();
    loadedTypeLibs = new Hashtable();
    
    importer = new Importer(this, loadedAssemblies);
    exporter = new Exporter(this, loadedTypeLibs);
    registrar = new Registrar(this, loadedAssemblies);
    comPlusInstaller = new ComPlusInstaller(this);
  }

  // Clean up any resources being used
  protected override void Dispose( bool disposing )
  {
    if (disposing)
    {
      // Abort the background thread
      if (backgroundThread != null)
        backgroundThread.Abort();

      if (components != null) 
      {
        components.Dispose();
      }
    }
    base.Dispose(disposing);
  }

  #region Windows Form Designer generated code
  //
  // Required method for Designer support - do not modify
  // the contents of this method with the code editor
  //
  private void InitializeComponent()
  {
    this.checkRegisteredTLBs = new System.Windows.Forms.CheckBox();
    this.checkArray = new System.Windows.Forms.CheckBox();
    this.textKeyContainer = new System.Windows.Forms.TextBox();
    this.groupRegister = new System.Windows.Forms.GroupBox();
    this.checkCodebase = new System.Windows.Forms.CheckBox();
    this.radioUnregister = new System.Windows.Forms.RadioButton();
    this.radioRegister = new System.Windows.Forms.RadioButton();
    this.checkVersion = new System.Windows.Forms.CheckBox();
    this.radioPublicKey = new System.Windows.Forms.RadioButton();
    this.comPlusPage = new System.Windows.Forms.TabPage();
    this.comPlusRegisterButton = new System.Windows.Forms.Button();
    this.groupComPlusTarget = new System.Windows.Forms.GroupBox();
    this.textPartitionName = new System.Windows.Forms.TextBox();
    this.textAppName = new System.Windows.Forms.TextBox();
    this.checkPartitionName = new System.Windows.Forms.CheckBox();
    this.checkAppName = new System.Windows.Forms.CheckBox();
    this.textTypeLibName = new System.Windows.Forms.TextBox();
    this.checkComponentsOnly = new System.Windows.Forms.CheckBox();
    this.checkNoReconfigure = new System.Windows.Forms.CheckBox();
    this.checkExistingTLB = new System.Windows.Forms.CheckBox();
    this.checkTypeLibName = new System.Windows.Forms.CheckBox();
    this.groupComPlusAction = new System.Windows.Forms.GroupBox();
    this.radioExistingApp = new System.Windows.Forms.RadioButton();
    this.radioUninstall = new System.Windows.Forms.RadioButton();
    this.radioCreate = new System.Windows.Forms.RadioButton();
    this.radioFindOrCreate = new System.Windows.Forms.RadioButton();
    this.registerNewButton = new System.Windows.Forms.Button();
    this.checkDelaySign = new System.Windows.Forms.CheckBox();
    this.checkLibName = new System.Windows.Forms.CheckBox();
    this.checkHelpFile = new System.Windows.Forms.CheckBox();
    this.textNames = new System.Windows.Forms.TextBox();
    this.openFileDialog1 = new System.Windows.Forms.OpenFileDialog();
    this.textLibName = new System.Windows.Forms.TextBox();
    this.importSaveButton = new System.Windows.Forms.Button();
    this.textVersion = new System.Windows.Forms.TextBox();
    this.registerPage = new System.Windows.Forms.TabPage();
    this.registerSelectedButton = new System.Windows.Forms.Button();
    this.checkPIA = new System.Windows.Forms.CheckBox();
    this.tabControl1 = new System.Windows.Forms.TabControl();
    this.importPage = new System.Windows.Forms.TabPage();
    this.importOpenButton = new System.Windows.Forms.Button();
    this.textNamespace = new System.Windows.Forms.TextBox();
    this.checkNamespace = new System.Windows.Forms.CheckBox();
    this.checkSign = new System.Windows.Forms.CheckBox();
    this.checkUnsafe = new System.Windows.Forms.CheckBox();
    this.groupSignAssembly = new System.Windows.Forms.GroupBox();
    this.radioKeyContainer = new System.Windows.Forms.RadioButton();
    this.radioKeyPair = new System.Windows.Forms.RadioButton();
    this.exportPage = new System.Windows.Forms.TabPage();
    this.exportNamesLabel = new System.Windows.Forms.Label();
    this.textHelpFile = new System.Windows.Forms.TextBox();
    this.textFilename = new System.Windows.Forms.TextBox();
    this.checkFilename = new System.Windows.Forms.CheckBox();
    this.checkNames = new System.Windows.Forms.CheckBox();
    this.exportSaveButton = new System.Windows.Forms.Button();
    this.exportOpenButton = new System.Windows.Forms.Button();
    this.treeView1 = new System.Windows.Forms.TreeView();
    this.progressBar1 = new System.Windows.Forms.ProgressBar();
    this.groupRegister.SuspendLayout();
    this.comPlusPage.SuspendLayout();
    this.groupComPlusTarget.SuspendLayout();
    this.groupComPlusAction.SuspendLayout();
    this.registerPage.SuspendLayout();
    this.tabControl1.SuspendLayout();
    this.importPage.SuspendLayout();
    this.groupSignAssembly.SuspendLayout();
    this.exportPage.SuspendLayout();
    this.SuspendLayout();
    // 
    // checkRegisteredTLBs
    // 
    this.checkRegisteredTLBs.Location = new System.Drawing.Point(8, 80);
    this.checkRegisteredTLBs.Name = "checkRegisteredTLBs";
    this.checkRegisteredTLBs.Size = new System.Drawing.Size(232, 16);
    this.checkRegisteredTLBs.TabIndex = 8;
    this.checkRegisteredTLBs.Text = "Reference only registered type libraries";
    // 
    // checkArray
    // 
    this.checkArray.Location = new System.Drawing.Point(8, 56);
    this.checkArray.Name = "checkArray";
    this.checkArray.Size = new System.Drawing.Size(216, 16);
    this.checkArray.TabIndex = 4;
    this.checkArray.Text = "Import SAFEARRAY as System.Array";
    // 
    // textKeyContainer
    // 
    this.textKeyContainer.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.textKeyContainer.Enabled = false;
    this.textKeyContainer.Location = new System.Drawing.Point(128, 56);
    this.textKeyContainer.Name = "textKeyContainer";
    this.textKeyContainer.Size = new System.Drawing.Size(104, 20);
    this.textKeyContainer.TabIndex = 11;
    this.textKeyContainer.Text = "";
    // 
    // groupRegister
    // 
    this.groupRegister.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                              this.checkCodebase,
                                                                              this.radioUnregister,
                                                                              this.radioRegister});
    this.groupRegister.Location = new System.Drawing.Point(8, 8);
    this.groupRegister.Name = "groupRegister";
    this.groupRegister.Size = new System.Drawing.Size(248, 80);
    this.groupRegister.TabIndex = 4;
    this.groupRegister.TabStop = false;
    this.groupRegister.Text = "Action";
    // 
    // checkCodebase
    // 
    this.checkCodebase.Location = new System.Drawing.Point(88, 16);
    this.checkCodebase.Name = "checkCodebase";
    this.checkCodebase.Size = new System.Drawing.Size(136, 16);
    this.checkCodebase.TabIndex = 1;
    this.checkCodebase.Text = "Use a codebase";
    // 
    // radioUnregister
    // 
    this.radioUnregister.Location = new System.Drawing.Point(8, 40);
    this.radioUnregister.Name = "radioUnregister";
    this.radioUnregister.Size = new System.Drawing.Size(104, 16);
    this.radioUnregister.TabIndex = 2;
    this.radioUnregister.Text = "Unregister";
    this.radioUnregister.CheckedChanged += new System.EventHandler(this.radioUnregister_CheckedChanged);
    // 
    // radioRegister
    // 
    this.radioRegister.Checked = true;
    this.radioRegister.Location = new System.Drawing.Point(8, 16);
    this.radioRegister.Name = "radioRegister";
    this.radioRegister.Size = new System.Drawing.Size(104, 16);
    this.radioRegister.TabIndex = 0;
    this.radioRegister.TabStop = true;
    this.radioRegister.Text = "Register";
    this.radioRegister.CheckedChanged += new System.EventHandler(this.radioRegister_CheckedChanged);
    // 
    // checkVersion
    // 
    this.checkVersion.Location = new System.Drawing.Point(8, 32);
    this.checkVersion.Name = "checkVersion";
    this.checkVersion.Size = new System.Drawing.Size(184, 16);
    this.checkVersion.TabIndex = 2;
    this.checkVersion.Text = "Customize version";
    this.checkVersion.CheckedChanged += new System.EventHandler(this.checkVersion_CheckedChanged);
    // 
    // radioPublicKey
    // 
    this.radioPublicKey.Enabled = false;
    this.radioPublicKey.Location = new System.Drawing.Point(16, 40);
    this.radioPublicKey.Name = "radioPublicKey";
    this.radioPublicKey.Size = new System.Drawing.Size(184, 16);
    this.radioPublicKey.TabIndex = 9;
    this.radioPublicKey.Text = "Use public key only from key file";
    // 
    // comPlusPage
    // 
    this.comPlusPage.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                            this.comPlusRegisterButton,
                                                                            this.groupComPlusTarget,
                                                                            this.textTypeLibName,
                                                                            this.checkComponentsOnly,
                                                                            this.checkNoReconfigure,
                                                                            this.checkExistingTLB,
                                                                            this.checkTypeLibName,
                                                                            this.groupComPlusAction});
    this.comPlusPage.Location = new System.Drawing.Point(4, 22);
    this.comPlusPage.Name = "comPlusPage";
    this.comPlusPage.Size = new System.Drawing.Size(264, 270);
    this.comPlusPage.TabIndex = 3;
    this.comPlusPage.Text = "COM+";
    // 
    // comPlusRegisterButton
    // 
    this.comPlusRegisterButton.FlatStyle = System.Windows.Forms.FlatStyle.Flat;
    this.comPlusRegisterButton.Location = new System.Drawing.Point(96, 240);
    this.comPlusRegisterButton.Name = "comPlusRegisterButton";
    this.comPlusRegisterButton.Size = new System.Drawing.Size(160, 24);
    this.comPlusRegisterButton.TabIndex = 13;
    this.comPlusRegisterButton.Text = "Register New Assembly...";
    this.comPlusRegisterButton.Click += new System.EventHandler(this.comPlusRegisterButton_Click);
    // 
    // groupComPlusTarget
    // 
    this.groupComPlusTarget.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                                   this.textPartitionName,
                                                                                   this.textAppName,
                                                                                   this.checkPartitionName,
                                                                                   this.checkAppName});
    this.groupComPlusTarget.Location = new System.Drawing.Point(8, 88);
    this.groupComPlusTarget.Name = "groupComPlusTarget";
    this.groupComPlusTarget.Size = new System.Drawing.Size(248, 64);
    this.groupComPlusTarget.TabIndex = 3;
    this.groupComPlusTarget.TabStop = false;
    this.groupComPlusTarget.Text = "Target Application";
    // 
    // textPartitionName
    // 
    this.textPartitionName.Enabled = false;
    this.textPartitionName.Location = new System.Drawing.Point(128, 40);
    this.textPartitionName.Name = "textPartitionName";
    this.textPartitionName.Size = new System.Drawing.Size(112, 20);
    this.textPartitionName.TabIndex = 7;
    this.textPartitionName.Text = "";
    // 
    // textAppName
    // 
    this.textAppName.Enabled = false;
    this.textAppName.Location = new System.Drawing.Point(128, 16);
    this.textAppName.Name = "textAppName";
    this.textAppName.Size = new System.Drawing.Size(112, 20);
    this.textAppName.TabIndex = 5;
    this.textAppName.Text = "";
    // 
    // checkPartitionName
    // 
    this.checkPartitionName.Location = new System.Drawing.Point(8, 40);
    this.checkPartitionName.Name = "checkPartitionName";
    this.checkPartitionName.Size = new System.Drawing.Size(232, 16);
    this.checkPartitionName.TabIndex = 6;
    this.checkPartitionName.Text = "Customize partition";
    this.checkPartitionName.CheckedChanged += new System.EventHandler(this.checkPartitionName_CheckedChanged);
    // 
    // checkAppName
    // 
    this.checkAppName.Location = new System.Drawing.Point(8, 16);
    this.checkAppName.Name = "checkAppName";
    this.checkAppName.Size = new System.Drawing.Size(232, 16);
    this.checkAppName.TabIndex = 4;
    this.checkAppName.Text = "Customize name";
    this.checkAppName.CheckedChanged += new System.EventHandler(this.checkAppName_CheckedChanged);
    // 
    // textTypeLibName
    // 
    this.textTypeLibName.Enabled = false;
    this.textTypeLibName.Location = new System.Drawing.Point(40, 168);
    this.textTypeLibName.Name = "textTypeLibName";
    this.textTypeLibName.Size = new System.Drawing.Size(208, 20);
    this.textTypeLibName.TabIndex = 9;
    this.textTypeLibName.Text = "";
    // 
    // checkComponentsOnly
    // 
    this.checkComponentsOnly.Location = new System.Drawing.Point(16, 224);
    this.checkComponentsOnly.Name = "checkComponentsOnly";
    this.checkComponentsOnly.Size = new System.Drawing.Size(232, 16);
    this.checkComponentsOnly.TabIndex = 12;
    this.checkComponentsOnly.Text = "Configure components only";
    // 
    // checkNoReconfigure
    // 
    this.checkNoReconfigure.Location = new System.Drawing.Point(16, 208);
    this.checkNoReconfigure.Name = "checkNoReconfigure";
    this.checkNoReconfigure.Size = new System.Drawing.Size(240, 16);
    this.checkNoReconfigure.TabIndex = 11;
    this.checkNoReconfigure.Text = "Don\'t reconfigure existing target application";
    // 
    // checkExistingTLB
    // 
    this.checkExistingTLB.Location = new System.Drawing.Point(16, 192);
    this.checkExistingTLB.Name = "checkExistingTLB";
    this.checkExistingTLB.Size = new System.Drawing.Size(232, 16);
    this.checkExistingTLB.TabIndex = 10;
    this.checkExistingTLB.Text = "Expect existing type library";
    // 
    // checkTypeLibName
    // 
    this.checkTypeLibName.Location = new System.Drawing.Point(16, 152);
    this.checkTypeLibName.Name = "checkTypeLibName";
    this.checkTypeLibName.Size = new System.Drawing.Size(232, 16);
    this.checkTypeLibName.TabIndex = 8;
    this.checkTypeLibName.Text = "Customize type library filename";
    this.checkTypeLibName.CheckedChanged += new System.EventHandler(this.checkTypeLibName_CheckedChanged);
    // 
    // groupComPlusAction
    // 
    this.groupComPlusAction.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                                   this.radioExistingApp,
                                                                                   this.radioUninstall,
                                                                                   this.radioCreate,
                                                                                   this.radioFindOrCreate});
    this.groupComPlusAction.Location = new System.Drawing.Point(8, 0);
    this.groupComPlusAction.Name = "groupComPlusAction";
    this.groupComPlusAction.Size = new System.Drawing.Size(248, 88);
    this.groupComPlusAction.TabIndex = 0;
    this.groupComPlusAction.TabStop = false;
    this.groupComPlusAction.Text = "Action";
    // 
    // radioExistingApp
    // 
    this.radioExistingApp.Location = new System.Drawing.Point(8, 64);
    this.radioExistingApp.Name = "radioExistingApp";
    this.radioExistingApp.Size = new System.Drawing.Size(200, 16);
    this.radioExistingApp.TabIndex = 3;
    this.radioExistingApp.Text = "Use existing application";
    // 
    // radioUninstall
    // 
    this.radioUninstall.Location = new System.Drawing.Point(8, 48);
    this.radioUninstall.Name = "radioUninstall";
    this.radioUninstall.Size = new System.Drawing.Size(200, 16);
    this.radioUninstall.TabIndex = 2;
    this.radioUninstall.Text = "Uninstall target application";
    this.radioUninstall.CheckedChanged += new System.EventHandler(this.radioUninstall_CheckedChanged);
    // 
    // radioCreate
    // 
    this.radioCreate.Location = new System.Drawing.Point(8, 32);
    this.radioCreate.Name = "radioCreate";
    this.radioCreate.Size = new System.Drawing.Size(200, 16);
    this.radioCreate.TabIndex = 1;
    this.radioCreate.Text = "Create the target application";
    // 
    // radioFindOrCreate
    // 
    this.radioFindOrCreate.Checked = true;
    this.radioFindOrCreate.Location = new System.Drawing.Point(8, 16);
    this.radioFindOrCreate.Name = "radioFindOrCreate";
    this.radioFindOrCreate.Size = new System.Drawing.Size(200, 16);
    this.radioFindOrCreate.TabIndex = 0;
    this.radioFindOrCreate.TabStop = true;
    this.radioFindOrCreate.Text = "Find or create target application";
    // 
    // registerNewButton
    // 
    this.registerNewButton.FlatStyle = System.Windows.Forms.FlatStyle.Flat;
    this.registerNewButton.Location = new System.Drawing.Point(96, 208);
    this.registerNewButton.Name = "registerNewButton";
    this.registerNewButton.Size = new System.Drawing.Size(160, 24);
    this.registerNewButton.TabIndex = 3;
    this.registerNewButton.Text = "Register New Assembly...";
    this.registerNewButton.Click += new System.EventHandler(this.registerNewButton_Click);
    // 
    // checkDelaySign
    // 
    this.checkDelaySign.Enabled = false;
    this.checkDelaySign.Location = new System.Drawing.Point(16, 80);
    this.checkDelaySign.Name = "checkDelaySign";
    this.checkDelaySign.Size = new System.Drawing.Size(96, 16);
    this.checkDelaySign.TabIndex = 12;
    this.checkDelaySign.Text = "Delay Sign";
    // 
    // checkLibName
    // 
    this.checkLibName.Location = new System.Drawing.Point(8, 32);
    this.checkLibName.Name = "checkLibName";
    this.checkLibName.Size = new System.Drawing.Size(144, 16);
    this.checkLibName.TabIndex = 2;
    this.checkLibName.Text = "Customize library name";
    this.checkLibName.CheckedChanged += new System.EventHandler(this.checkLibName_CheckedChanged);
    // 
    // checkHelpFile
    // 
    this.checkHelpFile.Location = new System.Drawing.Point(8, 56);
    this.checkHelpFile.Name = "checkHelpFile";
    this.checkHelpFile.Size = new System.Drawing.Size(144, 16);
    this.checkHelpFile.TabIndex = 4;
    this.checkHelpFile.Text = "Associate a help file";
    this.checkHelpFile.CheckedChanged += new System.EventHandler(this.checkHelpFile_CheckedChanged);
    // 
    // textNames
    // 
    this.textNames.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.textNames.Enabled = false;
    this.textNames.Location = new System.Drawing.Point(8, 144);
    this.textNames.Name = "textNames";
    this.textNames.Size = new System.Drawing.Size(248, 20);
    this.textNames.TabIndex = 7;
    this.textNames.Text = "";
    // 
    // textLibName
    // 
    this.textLibName.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.textLibName.Enabled = false;
    this.textLibName.Location = new System.Drawing.Point(144, 32);
    this.textLibName.Name = "textLibName";
    this.textLibName.Size = new System.Drawing.Size(112, 20);
    this.textLibName.TabIndex = 3;
    this.textLibName.Text = "";
    // 
    // importSaveButton
    // 
    this.importSaveButton.Enabled = false;
    this.importSaveButton.FlatStyle = System.Windows.Forms.FlatStyle.Flat;
    this.importSaveButton.Location = new System.Drawing.Point(144, 240);
    this.importSaveButton.Name = "importSaveButton";
    this.importSaveButton.Size = new System.Drawing.Size(112, 24);
    this.importSaveButton.TabIndex = 14;
    this.importSaveButton.Text = "Save Assembly";
    this.importSaveButton.Click += new System.EventHandler(this.importSaveButton_Click);
    // 
    // textVersion
    // 
    this.textVersion.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.textVersion.Enabled = false;
    this.textVersion.Location = new System.Drawing.Point(144, 32);
    this.textVersion.Name = "textVersion";
    this.textVersion.Size = new System.Drawing.Size(112, 20);
    this.textVersion.TabIndex = 3;
    this.textVersion.Text = "";
    // 
    // registerPage
    // 
    this.registerPage.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                             this.registerSelectedButton,
                                                                             this.registerNewButton,
                                                                             this.groupRegister});
    this.registerPage.Location = new System.Drawing.Point(4, 22);
    this.registerPage.Name = "registerPage";
    this.registerPage.Size = new System.Drawing.Size(264, 270);
    this.registerPage.TabIndex = 2;
    this.registerPage.Text = "Register";
    // 
    // registerSelectedButton
    // 
    this.registerSelectedButton.Enabled = false;
    this.registerSelectedButton.FlatStyle = System.Windows.Forms.FlatStyle.Flat;
    this.registerSelectedButton.Location = new System.Drawing.Point(96, 240);
    this.registerSelectedButton.Name = "registerSelectedButton";
    this.registerSelectedButton.Size = new System.Drawing.Size(160, 24);
    this.registerSelectedButton.TabIndex = 4;
    this.registerSelectedButton.Text = "Register Selected Item";
    this.registerSelectedButton.Click += new System.EventHandler(this.registerSelectedButton_Click);
    // 
    // checkPIA
    // 
    this.checkPIA.Location = new System.Drawing.Point(8, 104);
    this.checkPIA.Name = "checkPIA";
    this.checkPIA.Size = new System.Drawing.Size(208, 16);
    this.checkPIA.TabIndex = 6;
    this.checkPIA.Text = "Make a Primary Interop Assembly";
    // 
    // tabControl1
    // 
    this.tabControl1.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
      | System.Windows.Forms.AnchorStyles.Left);
    this.tabControl1.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                            this.importPage,
                                                                            this.exportPage,
                                                                            this.registerPage,
                                                                            this.comPlusPage});
    this.tabControl1.Location = new System.Drawing.Point(8, 8);
    this.tabControl1.Name = "tabControl1";
    this.tabControl1.SelectedIndex = 0;
    this.tabControl1.Size = new System.Drawing.Size(272, 296);
    this.tabControl1.TabIndex = 5;
    // 
    // importPage
    // 
    this.importPage.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                           this.importSaveButton,
                                                                           this.importOpenButton,
                                                                           this.textVersion,
                                                                           this.textNamespace,
                                                                           this.checkVersion,
                                                                           this.checkNamespace,
                                                                           this.checkSign,
                                                                           this.checkPIA,
                                                                           this.checkUnsafe,
                                                                           this.checkArray,
                                                                           this.groupSignAssembly});
    this.importPage.Location = new System.Drawing.Point(4, 22);
    this.importPage.Name = "importPage";
    this.importPage.Size = new System.Drawing.Size(264, 270);
    this.importPage.TabIndex = 0;
    this.importPage.Text = "Import";
    // 
    // importOpenButton
    // 
    this.importOpenButton.FlatStyle = System.Windows.Forms.FlatStyle.Flat;
    this.importOpenButton.Location = new System.Drawing.Point(8, 240);
    this.importOpenButton.Name = "importOpenButton";
    this.importOpenButton.Size = new System.Drawing.Size(128, 24);
    this.importOpenButton.TabIndex = 13;
    this.importOpenButton.Text = "Import Type Library...";
    this.importOpenButton.Click += new System.EventHandler(this.importOpenButton_Click);
    // 
    // textNamespace
    // 
    this.textNamespace.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.textNamespace.Enabled = false;
    this.textNamespace.Location = new System.Drawing.Point(144, 8);
    this.textNamespace.Name = "textNamespace";
    this.textNamespace.Size = new System.Drawing.Size(112, 20);
    this.textNamespace.TabIndex = 1;
    this.textNamespace.Text = "";
    // 
    // checkNamespace
    // 
    this.checkNamespace.Location = new System.Drawing.Point(8, 8);
    this.checkNamespace.Name = "checkNamespace";
    this.checkNamespace.Size = new System.Drawing.Size(144, 16);
    this.checkNamespace.TabIndex = 0;
    this.checkNamespace.Text = "Customize namespace";
    this.checkNamespace.CheckedChanged += new System.EventHandler(this.checkNamespace_CheckedChanged);
    // 
    // checkSign
    // 
    this.checkSign.Location = new System.Drawing.Point(8, 128);
    this.checkSign.Name = "checkSign";
    this.checkSign.Size = new System.Drawing.Size(117, 16);
    this.checkSign.TabIndex = 7;
    this.checkSign.Text = "Sign the Assembly";
    this.checkSign.CheckedChanged += new System.EventHandler(this.checkSign_CheckedChanged);
    // 
    // checkUnsafe
    // 
    this.checkUnsafe.Location = new System.Drawing.Point(8, 80);
    this.checkUnsafe.Name = "checkUnsafe";
    this.checkUnsafe.Size = new System.Drawing.Size(208, 16);
    this.checkUnsafe.TabIndex = 5;
    this.checkUnsafe.Text = "Create unsafe interfaces";
    // 
    // groupSignAssembly
    // 
    this.groupSignAssembly.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.groupSignAssembly.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                                  this.checkDelaySign,
                                                                                  this.textKeyContainer,
                                                                                  this.radioKeyContainer,
                                                                                  this.radioPublicKey,
                                                                                  this.radioKeyPair});
    this.groupSignAssembly.Location = new System.Drawing.Point(14, 128);
    this.groupSignAssembly.Name = "groupSignAssembly";
    this.groupSignAssembly.Size = new System.Drawing.Size(242, 104);
    this.groupSignAssembly.TabIndex = 4;
    this.groupSignAssembly.TabStop = false;
    // 
    // radioKeyContainer
    // 
    this.radioKeyContainer.Enabled = false;
    this.radioKeyContainer.Location = new System.Drawing.Point(16, 56);
    this.radioKeyContainer.Name = "radioKeyContainer";
    this.radioKeyContainer.Size = new System.Drawing.Size(120, 16);
    this.radioKeyContainer.TabIndex = 10;
    this.radioKeyContainer.Text = "Use key container";
    // 
    // radioKeyPair
    // 
    this.radioKeyPair.Checked = true;
    this.radioKeyPair.Enabled = false;
    this.radioKeyPair.Location = new System.Drawing.Point(16, 24);
    this.radioKeyPair.Name = "radioKeyPair";
    this.radioKeyPair.Size = new System.Drawing.Size(176, 16);
    this.radioKeyPair.TabIndex = 8;
    this.radioKeyPair.TabStop = true;
    this.radioKeyPair.Text = "Use key pair from key file";
    // 
    // exportPage
    // 
    this.exportPage.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                           this.exportNamesLabel,
                                                                           this.textNames,
                                                                           this.textHelpFile,
                                                                           this.checkHelpFile,
                                                                           this.textLibName,
                                                                           this.checkLibName,
                                                                           this.textFilename,
                                                                           this.checkFilename,
                                                                           this.checkRegisteredTLBs,
                                                                           this.checkNames,
                                                                           this.exportSaveButton,
                                                                           this.exportOpenButton});
    this.exportPage.Location = new System.Drawing.Point(4, 22);
    this.exportPage.Name = "exportPage";
    this.exportPage.Size = new System.Drawing.Size(264, 270);
    this.exportPage.TabIndex = 1;
    this.exportPage.Text = "Export";
    // 
    // exportNamesLabel
    // 
    this.exportNamesLabel.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.exportNamesLabel.FlatStyle = System.Windows.Forms.FlatStyle.System;
    this.exportNamesLabel.Location = new System.Drawing.Point(8, 168);
    this.exportNamesLabel.Name = "exportNamesLabel";
    this.exportNamesLabel.Size = new System.Drawing.Size(248, 16);
    this.exportNamesLabel.TabIndex = 11;
    this.exportNamesLabel.Text = "Identifiers should be space-delimited";
    this.exportNamesLabel.TextAlign = System.Drawing.ContentAlignment.TopCenter;
    // 
    // textHelpFile
    // 
    this.textHelpFile.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.textHelpFile.Enabled = false;
    this.textHelpFile.Location = new System.Drawing.Point(144, 56);
    this.textHelpFile.Name = "textHelpFile";
    this.textHelpFile.Size = new System.Drawing.Size(112, 20);
    this.textHelpFile.TabIndex = 5;
    this.textHelpFile.Text = "";
    // 
    // textFilename
    // 
    this.textFilename.Anchor = ((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.textFilename.Enabled = false;
    this.textFilename.Location = new System.Drawing.Point(144, 8);
    this.textFilename.Name = "textFilename";
    this.textFilename.Size = new System.Drawing.Size(112, 20);
    this.textFilename.TabIndex = 1;
    this.textFilename.Text = "";
    // 
    // checkFilename
    // 
    this.checkFilename.Location = new System.Drawing.Point(8, 8);
    this.checkFilename.Name = "checkFilename";
    this.checkFilename.Size = new System.Drawing.Size(144, 16);
    this.checkFilename.TabIndex = 0;
    this.checkFilename.Text = "Customize filename";
    this.checkFilename.CheckedChanged += new System.EventHandler(this.checkFilename_CheckedChanged);
    // 
    // checkNames
    // 
    this.checkNames.Location = new System.Drawing.Point(8, 128);
    this.checkNames.Name = "checkNames";
    this.checkNames.Size = new System.Drawing.Size(248, 16);
    this.checkNames.TabIndex = 6;
    this.checkNames.Text = "Use the specified casing for these identifiers:";
    this.checkNames.CheckedChanged += new System.EventHandler(this.checkNames_CheckedChanged);
    // 
    // exportSaveButton
    // 
    this.exportSaveButton.Enabled = false;
    this.exportSaveButton.FlatStyle = System.Windows.Forms.FlatStyle.Flat;
    this.exportSaveButton.Location = new System.Drawing.Point(144, 240);
    this.exportSaveButton.Name = "exportSaveButton";
    this.exportSaveButton.Size = new System.Drawing.Size(112, 24);
    this.exportSaveButton.TabIndex = 10;
    this.exportSaveButton.Text = "Save Type Library";
    this.exportSaveButton.Click += new System.EventHandler(this.exportSaveButton_Click);
    // 
    // exportOpenButton
    // 
    this.exportOpenButton.FlatStyle = System.Windows.Forms.FlatStyle.Flat;
    this.exportOpenButton.Location = new System.Drawing.Point(8, 240);
    this.exportOpenButton.Name = "exportOpenButton";
    this.exportOpenButton.Size = new System.Drawing.Size(128, 24);
    this.exportOpenButton.TabIndex = 9;
    this.exportOpenButton.Text = "Export Type Library...";
    this.exportOpenButton.Click += new System.EventHandler(this.exportOpenButton_Click);
    // 
    // treeView1
    // 
    this.treeView1.Anchor = (((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
      | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.treeView1.BorderStyle = System.Windows.Forms.BorderStyle.None;
    this.treeView1.ImageIndex = -1;
    this.treeView1.Location = new System.Drawing.Point(288, 8);
    this.treeView1.Name = "treeView1";
    this.treeView1.SelectedImageIndex = -1;
    this.treeView1.Size = new System.Drawing.Size(336, 296);
    this.treeView1.TabIndex = 4;
    this.treeView1.BeforeSelect += new System.Windows.Forms.TreeViewCancelEventHandler(this.treeView1_BeforeSelect);
    // 
    // progressBar1
    // 
    this.progressBar1.Anchor = ((System.Windows.Forms.AnchorStyles.Bottom | System.Windows.Forms.AnchorStyles.Left) 
      | System.Windows.Forms.AnchorStyles.Right);
    this.progressBar1.Location = new System.Drawing.Point(8, 312);
    this.progressBar1.Maximum = 0;
    this.progressBar1.Name = "progressBar1";
    this.progressBar1.Size = new System.Drawing.Size(616, 16);
    this.progressBar1.Step = 1;
    this.progressBar1.TabIndex = 6;
    // 
    // MainForm
    // 
    this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
    this.ClientSize = new System.Drawing.Size(632, 333);
    this.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                this.progressBar1,
                                                                this.tabControl1,
                                                                this.treeView1});
    this.Name = "MainForm";
    this.Text = "Interactive Interop Tool";
    this.groupRegister.ResumeLayout(false);
    this.comPlusPage.ResumeLayout(false);
    this.groupComPlusTarget.ResumeLayout(false);
    this.groupComPlusAction.ResumeLayout(false);
    this.registerPage.ResumeLayout(false);
    this.tabControl1.ResumeLayout(false);
    this.importPage.ResumeLayout(false);
    this.groupSignAssembly.ResumeLayout(false);
    this.exportPage.ResumeLayout(false);
    this.ResumeLayout(false);

  }
  #endregion

  // Implementation of IToolClient.SendMessage
  public void SendMessage(MessageTypes type, string message, object data)
  {
    // Call ActOnMessage on the main thread since it updates the TreeView
    Invoke(new MessageDelegate(ActOnMessage), 
      new object[]{type, message, data});
  }

  // Implementation of IToolClient.GetTypeLibReference
  public string GetTypeLibReference(Guid typeLibGuid, 
    string typeLibDisplayName)
  {
    // Display a MessageBox asking the user what to do next
    if (MessageBox.Show("The type library you're importing references " +
      "the type library '" + typeLibDisplayName + "' with LIBID " +
      typeLibGuid + ".  Do you want to select an assembly or type " +
      "library to satisfy this reference?  If you choose 'No', the " +
      "registered type library will be automatically imported with a " +
      "default assembly name, namespace, and version.",
      "Type Library Reference", MessageBoxButtons.YesNo, 
      MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == 
      DialogResult.Yes)
    {
      // The user chose to select a type library or assembly, so prepare
      // the OpenFileDialog
      openFileDialog1.Title = "Select an Assembly or Type Library";
      openFileDialog1.Filter = "Assemblies and Type Libraries (*.*)|*.*";
        
      if (openFileDialog1.ShowDialog() == DialogResult.OK)
      {
        // Return the chosen filename
        return openFileDialog1.FileName;
      }
    }

    // Stay with the default reference
    return null;
  }

  // Implementation of IToolClient.GetAssemblyReference
  public string GetAssemblyReference(string assemblyDisplayName)
  {
    // Display a MessageBox asking the user what to do next
    if (MessageBox.Show("The assembly you're exporting references the " +
      "assembly '" + assemblyDisplayName + "'.  Do you want to select " +
      "a type library or an assembly to satisfy this reference?  If you" +
      " choose 'No', the dependent assembly that's been located will be" +
      " automatically exported with a default filename, library name, " +
      "and help file.", "Assembly Reference", MessageBoxButtons.YesNo, 
      MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == 
      DialogResult.Yes)
    {
      // The user chose to select a type library or assembly, so prepare
      // the OpenFileDialog
      openFileDialog1.Title = "Select a Type Library or Assembly";
      openFileDialog1.Filter = "Type Libraries and Assemblies (*.*)|*.*";
        
      if (openFileDialog1.ShowDialog() == DialogResult.OK)
      {
        // Return the chosen filename
        return openFileDialog1.FileName;
      }
    }

    // Stay with the default reference
    return null;
  }

  // Implementation of IToolClient.AskYesNoQuestion
  public bool AskYesNoQuestion(string question, bool defaultIsYes)
  {
    // Select whether the default button is "Yes" or "No"
    MessageBoxDefaultButton defaultButton = defaultIsYes ? 
      MessageBoxDefaultButton.Button1 : MessageBoxDefaultButton.Button2;

    // Display a MessageBox asking the question
    if (MessageBox.Show(question, "Question", MessageBoxButtons.YesNo,
      MessageBoxIcon.Question, defaultButton) == DialogResult.Yes)
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  // NOTE: Updating the progress bar and caption should really be
  //       done using Control.Invoke since it's being done from
  //       a separate thread.

  // Implementation of IToolClient.ProgressMaximum
  public int ProgressMaximum
  {
    get { return progressBar1.Maximum; }
    set { progressBar1.Maximum = value; }
  }

  // Implementation of IToolClient.Progress
  public int Progress
  {
    get { return progressBar1.Value; }
    set 
    { 
      progressBar1.Value = value; 
      // Besides setting the progress bar value, update the form's caption
      Text = "Interactive Interop Tool - Converted " + 
        progressBar1.Value + " of " + progressBar1.Maximum + " Types";
    }
  }

  // Called by SendMessage so the work is done on the main thread
  private void ActOnMessage(MessageTypes type, string message, 
    object data)
  {
    switch (type)
    {
      case MessageTypes.BeginFile:
        Cursor = Cursors.WaitCursor;
        CustomTreeNode node = new CustomTreeNode(message);
        if (data != null) node.FilenameForSave = data.ToString();
        node.Nodes.Add("Invalid References");
        node.Nodes.Add("Warnings");
        node.Nodes.Add("Notifications");
        treeView1.Nodes.Add(node);
        filesBeingProcessed.Push(node);
        break;
      case MessageTypes.EndFile:
        ((TreeNode)filesBeingProcessed.Peek()).Tag = data;
        if (data as UCOMITypeLib != null)
        {
          // Append the LIBID of the type library
          ((TreeNode)filesBeingProcessed.Peek()).Text += 
            " (" + Marshal.GetTypeLibGuid(data as UCOMITypeLib) + ")";
        }
        filesBeingProcessed.Pop();
        if (filesBeingProcessed.Count == 0)
          ResetGui();
        break;
      case MessageTypes.InvalidRef:
        ((TreeNode)filesBeingProcessed.Peek()).Nodes[0].Nodes.Add(message);
        break;
      case MessageTypes.Warning:
        ((TreeNode)filesBeingProcessed.Peek()).Nodes[1].Nodes.Add(message);
        break;
      case MessageTypes.Notification:
        ((TreeNode)filesBeingProcessed.Peek()).Nodes[2].Nodes.Add(message);
        break;
    }
  }

  // Restore the GUI to an initial state after work is completed or if
  // an error occurs
  private void ResetGui()
  {
    foreach (object file in filesBeingProcessed)
      ((TreeNode)file).Remove();
    filesBeingProcessed.Clear();
    Text = "Interactive Interop Tool";
    Cursor = Cursors.Default;
    progressBar1.Value = 0;
    progressBar1.Maximum = 0;
  }

  // Called when user clicks "Import Type Library..." on Import tab
  private void importOpenButton_Click(object sender, System.EventArgs e)
  {
    // Do the importing on the background thread
    backgroundThread = new Thread(new ThreadStart(Import));
    backgroundThread.IsBackground = true;
    backgroundThread.ApartmentState = ApartmentState.MTA;
    backgroundThread.Start();
    tabControl1.Enabled = false;
  }

  // Implementation of "Import Type Library..." running on 
  // the background thread
  private void Import()
  {
    try
    {
      ImportOptions options = new ImportOptions();
        
      string @namespace = null;
      Version version = null;

      // Fill in the namespace if checkNamespace is checked
      if (checkNamespace.Checked)
        @namespace = textNamespace.Text;

      // Fill in the version if checkVersion is checked
      if (checkVersion.Checked)
      {
        try
        {
          version = new Version(textVersion.Text);
        }
        catch (Exception ex)
        {
          MessageBox.Show("The version is invalid.  " + ex.Message, 
            "Invalid Version Number", MessageBoxButtons.OK, 
            MessageBoxIcon.Error);
          return;
        }
      }

      // Fill in the TypeLibImporterFlags values
      if (checkArray.Checked) 
        options.Flags |= TypeLibImporterFlags.SafeArrayAsSystemArray;
      if (checkUnsafe.Checked) 
        options.Flags |= TypeLibImporterFlags.UnsafeInterfaces;
      if (checkPIA.Checked) 
        options.Flags |= TypeLibImporterFlags.PrimaryInteropAssembly;

      // Prepare the OpenFileDialog to ask for a type library
      openFileDialog1.Title = "Select a Type Library";
      openFileDialog1.Filter = 
        "Type Libraries (*.olb;*.tlb;*.dll)|*.olb;*.tlb;*.dll|" +
        "Executable Files (*.exe;*.dll)|*.exe;*.dll|" +
        "ActiveX Controls (*.ocx)|*.ocx|" +
        "All Files (*.*)|*.*";

      if (openFileDialog1.ShowDialog() == DialogResult.OK)
      {
        string tlbFile = openFileDialog1.FileName;
        try
        {
          // Handle the strong name options if checkSign is checked
          if (checkSign.Checked)
          {
            options.UseKeyPairFromFile = radioKeyPair.Checked;

            // If either of these two are checked, we need to ask the user to
            // supply a filename containing the key data
            if (radioKeyPair.Checked || radioPublicKey.Checked)
            {
              // Prepare the OpenFileDialog
              openFileDialog1.Title = "Select a Key File to Sign the Assembly";
              openFileDialog1.Filter = "All Files (*.*)|*.*";
                
              if (openFileDialog1.ShowDialog() == DialogResult.OK)
              {
                // Set ImportOptions.KeyFile to the given filename
                options.KeyFile = openFileDialog1.FileName;
              }
            }

            if (radioKeyContainer.Checked)
              options.KeyContainer = textKeyContainer.Text;

            options.DelaySign = checkDelaySign.Checked;
          }

          // Now all the options are ready for import
          importer.Options = options;
          importer.OutputDirectory = Path.GetDirectoryName(tlbFile);
          // Use the namespace for both namespace and assembly name
          importer.Import(tlbFile, @namespace, 
            @namespace, version);
        }
        catch (Exception ex)
        {
          MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, 
            MessageBoxIcon.Error);
          ResetGui();
        }
      }
    }
    finally
    {
      tabControl1.Enabled = true;
    }
  }

  // Called when user clicks "Save Assembly" on Import tab
  private void importSaveButton_Click(object sender, System.EventArgs e)
  {
    AssemblyBuilder asmToSave = 
      treeView1.SelectedNode.Tag as AssemblyBuilder;

    // If the assembly is an existing Primary Interop Assembly,
    // we simply loaded it with Assembly.Load or Assembly.LoadFrom so
    // the object would not be an AssemblyBuilder.
    if (asmToSave == null)
    {
      MessageBox.Show("This assembly does not support being saved.",
        "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
      return;
    }

    // Display a warning and allow the user to cancel if a file 
    // with the same name already exists
    if (File.Exists(
      ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave))
    {
      if (MessageBox.Show("This assembly must be saved as '" +
        ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave + "'" +
        " but this file already exists.  Do you wish to continue and " +
        "overwrite the file?", "File Already Exists", 
        MessageBoxButtons.YesNo, MessageBoxIcon.Warning, 
        MessageBoxDefaultButton.Button2) == DialogResult.No)
        return;
    }

    try
    {
      // Save the assembly
      asmToSave.Save(asmToSave.GetName().Name + ".dll");
      MessageBox.Show("Successfully saved assembly as '" +
        ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave + "'.",
        "Save Successful", MessageBoxButtons.OK, 
        MessageBoxIcon.Information);
    }
    catch (Exception ex)
    {
      MessageBox.Show("Unable to save assembly.  " + ex.Message, "Error",
        MessageBoxButtons.OK, MessageBoxIcon.Error);
    }
  }

  // Called when user clicks "Export Type Library..." on Export tab
  private void exportOpenButton_Click(object sender, System.EventArgs e)
  {
    // Do the exporting on the background thread
    backgroundThread = new Thread(new ThreadStart(Export));
    backgroundThread.IsBackground = true;
    backgroundThread.ApartmentState = ApartmentState.MTA;
    backgroundThread.Start();
    tabControl1.Enabled = false;
  }

  // Implementation of "Export Type Library..." running on 
  // the background thread
  private void Export()
  {
    try
    {
      ExportOptions options = new ExportOptions();

      string filename = null;
      string libraryName = null;
      string helpFile = null;

      // Fill in the filename if checkFilename is checked
      if (checkFilename.Checked)
        filename = textFilename.Text;

      // Fill in the library name if checkLibName is checked
      if (checkLibName.Checked)
        libraryName = textLibName.Text;

      // Fill in the help file if checkHelpFile is checked
      if (checkHelpFile.Checked)
        helpFile = textHelpFile.Text;

      // Fill in the names file if checkNames is checked
      if (checkNames.Checked)
        options.Names = textNames.Text.Split();

      if (checkRegisteredTLBs.Checked)
        options.Flags = TypeLibExporterFlags.OnlyReferenceRegistered;

      // Prepare the OpenFileDialog for choosing an assembly
      openFileDialog1.Title = "Select an Assembly";
      openFileDialog1.Filter = "Assemblies (*.dll;*.exe)|*.dll;*.exe|" +
        "All Files (*.*)|*.*";

      if (openFileDialog1.ShowDialog() == DialogResult.OK)
      {
        try
        {
          exporter.Options = options;
          exporter.OutputDirectory = 
            Path.GetDirectoryName(openFileDialog1.FileName);
          exporter.Export(openFileDialog1.FileName, filename, 
            libraryName, helpFile);
        }
        catch (Exception ex)
        {
          MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, 
            MessageBoxIcon.Error);
          ResetGui();
        }
      }
    }
    finally
    {
      tabControl1.Enabled = true;
    }
  }

  // Called when user clicks "Save Type Library" on Export tab
  private void exportSaveButton_Click(object sender, System.EventArgs e)
  {
    ICreateTypeLib typeLibToSave = 
      treeView1.SelectedNode.Tag as ICreateTypeLib;

    // If the type library was embedded in an assembly, we simply 
    // loaded it with LoadTypeLibEx so the object 
    // would not support ICreateTypeLib.
    if (typeLibToSave == null)
    {
      MessageBox.Show("This type library does not support being saved.",
        "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
      return;
    }

    // Display a warning and allow the user to cancel if a file 
    // with the same name already exists
    if (File.Exists(
      ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave))
    {
      if (MessageBox.Show("This type library must be saved as '" + 
        ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave +
        "' but this file already exists.  Do you wish to continue and " +
        "overwrite the file?", "File Already Exists", 
        MessageBoxButtons.YesNo, MessageBoxIcon.Warning, 
        MessageBoxDefaultButton.Button2) == DialogResult.No)
        return;
    }

    try
    {
      // Save the type library
      typeLibToSave.SaveAllChanges();
      MessageBox.Show("Successfully saved type library as '" +
        ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave +
        "'.", "Save Successful", MessageBoxButtons.OK, 
        MessageBoxIcon.Information);
    }
    catch (Exception ex)
    {
      MessageBox.Show("Unable to save type library.  " + ex.Message, 
        "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
    }
  }

  // Called when user clicks "Register New Assembly..." on Register tab
  private void registerNewButton_Click(object sender, System.EventArgs e)
  {
    RegisterOptions options = new RegisterOptions();
    string result;

    // Set the codebase if it's checked and enabled
    if (checkCodebase.Checked && checkCodebase.Enabled)
      options.Flags = AssemblyRegistrationFlags.SetCodeBase;

    // Prepare the OpenFileDialog for choosing an assembly
    openFileDialog1.Title = "Select an Assembly";
    openFileDialog1.Filter = "Assemblies (*.dll;*.exe)|*.dll;*.exe|" +
      "All Files (*.*)|*.*";

    if (openFileDialog1.ShowDialog() == DialogResult.OK)
    {
      try
      {
        registrar.Options = options;

        // Either register or unregister the assembly based on the value
        // of radioRegister
        if (radioRegister.Checked)
          result = registrar.Register(openFileDialog1.FileName);
        else
          result = registrar.Unregister(openFileDialog1.FileName);

        MessageBox.Show(result, "Success", MessageBoxButtons.OK,
          MessageBoxIcon.Information);
      }
      catch (Exception ex)
      {
        MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, 
          MessageBoxIcon.Error);
      }
    }
  }

  // Called when user clicks "Register Selected Item" on Register tab
  private void registerSelectedButton_Click(object sender, 
    System.EventArgs e)
  {
    if (treeView1.SelectedNode.Tag as Assembly != null)
    {
      // Code for (un)registering an in-memory assembly
      RegisterOptions options = new RegisterOptions();
      string result;

      if (checkCodebase.Checked && checkCodebase.Enabled)
        options.Flags = AssemblyRegistrationFlags.SetCodeBase;

      registrar.Options = options;

      try
      {
        if (radioRegister.Checked)
        {
          result = registrar.Register(
            (Assembly)treeView1.SelectedNode.Tag);
        }
        else
        {
          result = registrar.Unregister(
            (Assembly)treeView1.SelectedNode.Tag);
        }

        MessageBox.Show(result, "Success", MessageBoxButtons.OK,
          MessageBoxIcon.Information);
      }
      catch (NotSupportedException ex)
      {
        MessageBox.Show(ex.Message + "  Make sure that the assembly " +
          "has been saved before attempting to " +
          (radioRegister.Checked ? "" : "un") + "register it.", 
          "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
      }
      catch (Exception ex)
      {
        MessageBox.Show(ex.Message, 
          "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
      }
    }
    else if (treeView1.SelectedNode.Tag as UCOMITypeLib != null)
    {
      if (radioRegister.Checked)
      {
        // Code for registering an in-memory type library
        try
        {
          Util.LoadTypeLibEx(
            ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave, 
            REGKIND.REGKIND_REGISTER);

          MessageBox.Show("Successfully registered the type library '" +
            ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave +
            "'.", "Success", MessageBoxButtons.OK,
            MessageBoxIcon.Information);
        }
        catch
        {
          MessageBox.Show("Unable to register the type library '" +
            ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave +
            "'.  Make sure that the type library has been saved " +
            "before attempting to register it.", "Error", 
            MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
      }
      else
      {
        // Code for unregistering an in-memory type library
        IntPtr attributesPtr = IntPtr.Zero;
        try
        {
          // UCOMITypeLib.GetLibAttr takes a System.IntPtr rather than a
          // TYPELIBATTR directly, so the followng code is necessary.
          ((UCOMITypeLib)treeView1.SelectedNode.Tag).GetLibAttr(
            out attributesPtr);
          TYPELIBATTR attributes = (TYPELIBATTR)Marshal.PtrToStructure(
            attributesPtr, typeof(TYPELIBATTR));
          Util.UnRegisterTypeLib(ref attributes.guid, 
            attributes.wMajorVerNum, attributes.wMinorVerNum, 
            attributes.lcid, attributes.syskind);
        }
        catch (COMException ex)
        {
          // TYPE_E_REGISTRYACCESS might be returned, which is a harmless 
          // error indicating that the type library 
          // is already unregistered
          if ((uint)ex.ErrorCode != 0x8002801C)
            throw;
        }
        catch (Exception ex)
        {
          MessageBox.Show("Unable to unregister the type library.  " + 
            ex.Message, "Error", MessageBoxButtons.OK, 
            MessageBoxIcon.Error);
        }
        finally
        {
          if (attributesPtr != IntPtr.Zero)
            ((UCOMITypeLib)treeView1.SelectedNode.Tag).ReleaseTLibAttr(
              attributesPtr);
        }

        MessageBox.Show("Successfully unregistered the type library '" +
          ((CustomTreeNode)treeView1.SelectedNode).FilenameForSave +
          "'.", "Success", MessageBoxButtons.OK,
          MessageBoxIcon.Information);
      }
    }
  }

  // Called when user clicks "Register New Assembly..." on COM+ tab
  private void comPlusRegisterButton_Click(object sender, 
    System.EventArgs e)
  {
    ComPlusInstallerOptions options = new ComPlusInstallerOptions();
    string result;

    // Handle all the options
    if (radioFindOrCreate.Checked)
      options.Flags |= InstallationFlags.FindOrCreateTargetApplication;
      
    if (radioCreate.Checked)
      options.Flags |= InstallationFlags.CreateTargetApplication;

    if (checkExistingTLB.Checked)
      options.Flags |= InstallationFlags.ExpectExistingTypeLib;

    if (checkTypeLibName.Checked)
      options.TypeLibName = textTypeLibName.Text;

    if (!checkNoReconfigure.Checked)
      options.Flags |= InstallationFlags.ReconfigureExistingApplication;

    if (checkComponentsOnly.Checked)
      options.Flags |= InstallationFlags.ConfigureComponentsOnly;

    if (checkAppName.Checked)
      options.ApplicationName = textAppName.Text;

    if (checkPartitionName.Checked)
      options.PartitionName = textPartitionName.Text;
    
    // Prepare the OpenFileDialog for choosing an assembly
    openFileDialog1.Title = "Select an Assembly";
    openFileDialog1.Filter = "Assemblies (*.dll;*.exe)|*.dll;*.exe|" +
      "All Files (*.*)|*.*";

    if (openFileDialog1.ShowDialog() == DialogResult.OK)
    {
      try
      {
        comPlusInstaller.Options = options;

        // Install or uninstall the assembly based on the value
        // of radioUninstall
        if (radioUninstall.Checked)
          result = comPlusInstaller.Install(openFileDialog1.FileName);
        else
          result = comPlusInstaller.Uninstall(openFileDialog1.FileName);

        MessageBox.Show(result, "Success", MessageBoxButtons.OK,
          MessageBoxIcon.Information);
      }
      catch (Exception ex)
      {
        MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, 
          MessageBoxIcon.Error);
      }
    }
  }

  // Enables/disables the appropriate buttons when items in the
  // TreeView are selected
  private void treeView1_BeforeSelect(object sender, 
    System.Windows.Forms.TreeViewCancelEventArgs e)
  {
    importSaveButton.Enabled = false;
    exportSaveButton.Enabled = false;
    registerSelectedButton.Enabled = false;

    if (e.Node.Tag as AssemblyBuilder != null)
    {
      // The node represents a dynamic assembly
      importSaveButton.Enabled = true;
      registerSelectedButton.Enabled = true;
    }
    else if (e.Node.Tag as ICreateTypeLib != null)
    {
      // The node represents a dynamic type library
      exportSaveButton.Enabled = true;
      registerSelectedButton.Enabled = true;
    }

    if (e.Node.Tag as Assembly != null)
    {
      // The node represents an assembly
      registerSelectedButton.Enabled = true;
    }
    else if (e.Node.Tag as UCOMITypeLib != null)
    {
      // The node represents a type library
      registerSelectedButton.Enabled = true;
    }
  }

  // For some CheckBox controls on the Import tab,
  // enable/disable corresponding text boxes

  private void checkNamespace_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (checkNamespace.Checked)
      textNamespace.Enabled = true;
    else
      textNamespace.Enabled = false;
  }

  private void checkSign_CheckedChanged(object sender, System.EventArgs e)
  {
    if (checkSign.Checked)
    {
      radioKeyPair.Enabled = true;
      radioPublicKey.Enabled = true;
      radioKeyContainer.Enabled = true;
      checkDelaySign.Enabled = true;
    }
    else
    {
      radioKeyPair.Enabled = false;
      radioPublicKey.Enabled = false;
      radioKeyContainer.Enabled = false;
      checkDelaySign.Enabled = false;
    }
  }

  private void radioKeyContainer_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (radioKeyContainer.Checked)
      textKeyContainer.Enabled = true;
    else
      textKeyContainer.Enabled = false;
  }

  private void checkVersion_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (checkVersion.Checked)
      textVersion.Enabled = true;
    else
      textVersion.Enabled = false;
  }

  // For some CheckBox controls on the Export tab,
  // enable/disable corresponding text boxes

  private void checkFilename_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (checkFilename.Checked)
      textFilename.Enabled = true;
    else
      textFilename.Enabled = false;
  }

  private void checkLibName_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (checkLibName.Checked)
      textLibName.Enabled = true;
    else
      textLibName.Enabled = false;
  }

  private void checkHelpFile_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (checkHelpFile.Checked)
      textHelpFile.Enabled = true;
    else
      textHelpFile.Enabled = false;
  }

  private void checkNames_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (checkNames.Checked)
      textNames.Enabled = true;
    else
      textNames.Enabled = false;
  }

  // For some CheckBox controls on the Register tab,
  // enable/disable corresponding text boxes

  private void radioRegister_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (radioRegister.Checked)
    {
      registerNewButton.Text = "Register New Assembly...";
      registerSelectedButton.Text = "Register Selected Item";
      checkCodebase.Enabled = true;
    }
  }

  private void radioUnregister_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (radioUnregister.Checked)
    {
      registerNewButton.Text = "Unregister New Assembly...";
      registerSelectedButton.Text = "Unregister Selected Item";
      checkCodebase.Enabled = false;
    }
  }

  // For some CheckBox controls on the COM+ tab,
  // enable/disable corresponding text boxes

  private void checkAppName_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (checkAppName.Checked)
      textAppName.Enabled = true;
    else
      textAppName.Enabled = false;
  }

  private void checkPartitionName_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (checkPartitionName.Checked)
      textPartitionName.Enabled = true;
    else
      textPartitionName.Enabled = false;
  }

  private void checkTypeLibName_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (checkTypeLibName.Checked)
      textTypeLibName.Enabled = true;
    else
      textTypeLibName.Enabled = false;
  }

  private void radioUninstall_CheckedChanged(object sender, 
    System.EventArgs e)
  {
    if (radioUninstall.Checked)
      comPlusRegisterButton.Text = "Unregister New Assembly...";
    else
      comPlusRegisterButton.Text = "Register New Assembly...";
  }
}