using DirectX;

public class AnimatedPicture : Picture
{
  // Number of frames in the animation
  protected int frames;       
  // Number of frames in one row of the input bitmap 
  protected int framesInRow;  
  // Width of each frame 
  protected int frameWidth;   
  // Height of each frame 
  protected int frameHeight;  
  // Number of times drawn before changing the frame.
  // Thus, this controls the speed of the animation.
  protected int frequency;    

  // Creates the picture from the bitmap.  
  // The bitmap contains each frame of the animation.
  public AnimatedPicture(string bitmapFile, bool hasTransparentColor,
    int numFrames, int rows, int frequency) 
    : base(bitmapFile, hasTransparentColor)
  {
    frames = numFrames;
    framesInRow = numFrames / rows;
    frameWidth = (size.Right - size.Left) / framesInRow;
    frameHeight = (size.Bottom - size.Top) / (numFrames / framesInRow);
    this.frequency = frequency;
  }

  // Returns the width of the picture
  public override int Width
  {
    // Return the width of just one frame, not the entire input bitmap
    get { return frameWidth; }
  }

  // Returns the height the picture
  public override int Height
  {
    // Return the height of just one frame, not the entire input bitmap
    get { return frameHeight; }
  }

  // Draw the picture with the top left corner at the
  // (x,y) coordinate.  The picture is clipped if part
  // of the picture is off the screen.
  public override void DrawAt(int x, int y)
  {
    RECT frame = new RECT();

    // Get the correct "slice" of the input bitmap based
    // on the current frame number.
    int frameNumber = (Util.Counter / frequency) % frames;
    frame.Left = (frameNumber % framesInRow) * frameWidth;
    frame.Top = (frameNumber / framesInRow) * frameHeight;
    frame.Right = frame.Left + frameWidth;
    frame.Bottom = frame.Top + frameHeight;

    RECT r = ClippedRectangle(ref x, ref y, frame);

    Util.DrawSurface(dxSurface, x, y, r, hasTransparentColor);
  }

  // Draw the picture with the top left corner at the
  // (x,y) coordinate, resized to the width and height given.
  // The picture is clipped if part of the picture is off
  // the screen (horizontally only).
  public override void ResizeAndDrawAt(
    int x, int y, int width, int height)
  {
    RECT frame = new RECT();

    // Get the correct "slice" of the input bitmap based
    // on the current frame number.
    int frameNumber = (Util.Counter / frequency) % frames;
    frame.Left = (frameNumber % framesInRow) * frameWidth;
    frame.Top = (frameNumber / framesInRow) * frameHeight;
    frame.Right = frame.Left + frameWidth;
    frame.Bottom = frame.Top + frameHeight;

    RECT r = ClippedRectangle(ref x, ref y, frame);

    RECT dest = new RECT();
    dest.Left = x;
    dest.Top = y;
    dest.Right = x + width;
    dest.Bottom = y + height;

    Util.ResizeAndDrawSurface(dxSurface, r, dest, hasTransparentColor);
  }
}