public class Envelope
{
  // Number of different speeds that envelopes may have
  private const int numSpeeds = 3;       
  // Frequency of the fastest envelopes.
  // The next fastest are 2 * frequencyUnit, then 3 * frequencyUnit, ...
  private const int frequencyUnit = 15;  
  // True if the envelope is visible
  public bool IsVisible;        
  // True if the character has stepped on the envelope 
  public bool IsInfected;       
  // True if the envelope has an attachment
  public bool HasAttachment;    
  // Size of the envelope, corresponding to the frame in 
  // the EnvelopePicture.  0 is the largest size.
  private int size;             

  // Number of frames between size changes
  private int shrinkFrequency;  
  // "ticks" since envelope became visible 
  private int timeVisible;      
  // Current horizontal location of the envelope (in pixels)
  private int locationX;        
  // Current vertical location of the envelope (in pixels)
  private int locationY;        
  // Initial horizontal location of the envelope (in pixels)
  private int initialX;         
  // Frequency of the layer containing the envelopes 
  private int layerFrequency;   

  // Creates the envelope at the initial location given
  public Envelope(int initialX, int initialY, int layerFrequency)
  {
    locationX = initialX;
    locationY = initialY;
    this.initialX = initialX;
    this.layerFrequency = layerFrequency;
    ChooseNewCharacteristics();
  }

  // Re-initializes the envelopes to the state when constructed
  public void Reset()
  {
    locationX = initialX;
    ChooseNewCharacteristics();
  }

  // Sets new characteristics for the envelope.  The values for
  // IsVisible, HasAttachment, and shrinkFrequency are chosen randomly.
  private void ChooseNewCharacteristics()
  {
    timeVisible = 0;
    size = 0;
    IsInfected = false;
    IsVisible = (Util.RandomNumber.Next(4) == 0);
    HasAttachment = (Util.RandomNumber.Next(20) == 0);
    shrinkFrequency = 
      (Util.RandomNumber.Next(numSpeeds) + 1) * frequencyUnit;
  }

  // Ensures that the envelope is visible and infected with no attachment.
  // The envelope also shrinks at the slowest speed.
  public void MakeVisibleAndInfected()
  {
    ChooseNewCharacteristics();
    shrinkFrequency = numSpeeds * frequencyUnit;
    IsInfected = true;
    IsVisible = true;
    HasAttachment = false;
  }

  // Draw the envelope in its current state.
  // This must be followed by a call to Util.RenderComplete().
  public void Render()
  {
    timeVisible++;

    // Move the envelope along with the scrolling layer it's on.
    if (Util.IsTimeToMove(layerFrequency))
      locationX--;

    // Make the envelope wrap around the screen.  When it reappears
    // on the right, it's a "new" envelope with new characteristics.
    if (locationX + Game.EnvelopePicture.Width < 0)
    {
      ChooseNewCharacteristics();
      locationX = Screen.Width + Game.EnvelopePadding - 1;
      return;
    }

    if (IsVisible)
    {
      // Make it the correct size
      if (timeVisible % (shrinkFrequency) == 0)
      {
        size++;
        if (size == AnimatedEnvelopePicture.NumFrames)
        {
          // The envelope is past its smallest size,
          // so make it invisible.
          IsVisible = false;
          return;
        }
      }

      Game.EnvelopePicture.DrawAt(locationX, locationY, size, IsInfected);

      // Draw an attachment, if it exists
      if (HasAttachment)
      {
        Game.BonusPicture.DrawAt(
          locationX + (Game.EnvelopePicture.Width/2) - 
            (Game.BonusPicture.Width / 2),
          locationY + (Game.EnvelopePicture.Height/2) - 
            (Game.BonusPicture.Height / 2)
        );
      }
    }
    else if (!Game.Player.IsDying)
    {
      // Choose new characteristics with a low chance of visibility,
      // since this is done every frame when its invisible.
      // This is only done when the player isn't in the process
      // of dying so it's not possible for an envelope to appear
      // under the falling character.
      ChooseNewCharacteristics();
      IsVisible = (Util.RandomNumber.Next(600) == 0);
    }
  }
}