using System;
using Microsoft.Win32;

public class Game : IDisposable
{
  // Number of envelopes in the grid horizontally
  public const int GridWidth = 9;
  // Number of envelopes in the grid vertically
  public const int GridHeight = 7;
  // Space between envelopes
  public const int EnvelopePadding = 6;
  // Space between the top of the screen and the topmost row of envelopes
  public const int TopSpace = 26;             
  // Space between the left edge of the screen and
  // the leftmost column of envelopes
  public const int LeftSpace = 3;             
  // Number of frames between each movement of the scrolling ground
  public const int ScrollFrequency = 2;
  // Space between digits in the ScorePicture
  public const int DigitPadding = 4;
  // Height of digits in the ScorePicture
  public const int DigitHeight = 44;            
  // Maximum width of the digits in the ScorePicture
  public const int MaxDigitWidth = 30;        
  // Width of each digit in the ScorePicture, 0 to 9
  public readonly int [] DigitWidths = 
    new int[]{27, 13, 28, 24, 30, 26, 27, 26, 27, 28};

  // The player
  public static Character Player;                        
  // The grid of envelopes
  public static Envelope [,] EnvelopeGrid;               
  // The picture for each bonus
  public static AnimatedPicture BonusPicture;
  // The picture for each envelope
  public static AnimatedEnvelopePicture EnvelopePicture;
  // The picture for the player
  public static AnimatedPicture PlayerPicture;             
  // The picture for the score
  public static NumberPicture ScorePicture;              

  // The first screen shown
  private static IntroScreen titleScreen;
  // The pause screen
  private static Screen pauseScreen;          
  // The screen shown after each life lost
  private static Screen tryAgainScreen;     

  // The scrolling layers
  private static Layer [] layers;           
  // The current state of the game 
  private static GameStates state;          
  // Number of lives remaining 
  private static int numLives;              
  // Number of bonuses remaining 
  private static int numBonuses;            
  // The current score 
  private static int score;                 
  // The high score, stored in the Windows Registry 
  private static int highScore;             

  // Sound made when the player moves
  private static Sound moveSound;           
  // Sound made when the player infects an envelope 
  private static Sound infectSound;         
  // Sound made when the player gets a bonus
  private static Sound getBonusSound;       
  // Sound made when the player uses a bonus
  private static Sound useBonusSound;       
  // Sound made when the player makes an invalid move 
  private static Sound invalidMoveSound;    
  // Sound made when the player loses a life 
  private static Sound lostLifeSound;       
  // Sound made when the player gains an extra life 
  private static Sound extraLifeSound;      

  // Initializes the pictures and sounds, and displays the first screen
  public Game(IntPtr hWnd)
  {
    // Initialize the Util helper class.
    Util.Initialize(hWnd);

    // Create the screens.
    titleScreen = new IntroScreen("title.bmp");
    pauseScreen = new Screen("paused.bmp");
    tryAgainScreen = new Screen("tryagain.bmp");

    // Create the pictures.
    EnvelopePicture = new AnimatedEnvelopePicture("envelope.bmp");
    BonusPicture = new AnimatedPicture("bonus.bmp", true, 21, 3, 2);
    PlayerPicture = new AnimatedPicture("character.bmp", true, 2, 1, 2);
    ScorePicture = new NumberPicture("digits.bmp", true, DigitWidths,
      MaxDigitWidth, DigitPadding);

    // Create the layers.
    layers = new Layer[4];
    layers[0] = new Layer("layer1.bmp", false, ScrollFrequency * 4);
    layers[1] = new Layer("layer2.bmp", true, ScrollFrequency * 2);
    layers[2] = new GroundLayer("layer3.bmp", true, ScrollFrequency);
    layers[3] = new FrontLayer("layer4.bmp");

    // Fill the grid with envelopes.
    EnvelopeGrid = new Envelope[GridWidth, GridHeight];
    for (int i = 0; i < GridWidth; i++)
      for (int j = 0; j < GridHeight; j++)
        EnvelopeGrid[i,j] = new Envelope(
          LeftSpace + i * (EnvelopePicture.Width + EnvelopePadding),
          TopSpace + j * (EnvelopePicture.Height + EnvelopePadding),
          layers[2].Frequency);

    // Create the sounds.
    moveSound = new Sound("move.wav");
    infectSound = new Sound("infect.wav");
    getBonusSound = new Sound("getbonus.wav");
    useBonusSound = new Sound("usebonus.wav");
    invalidMoveSound = new Sound("invalidmove.wav");
    lostLifeSound = new Sound("lostlife.wav");
    extraLifeSound = new Sound("extralife.wav");

    // Create the player.
    Player = new Character();

    highScore = -1;
    score = 0;
    State = GameStates.NotStarted;
  }

  // Disposes the game
  ~Game()
  {
    Dispose();
  }

  // Disposes the game
  public void Dispose()
  {
    // Since we're cleaning up now, suppress finalization
    GC.SuppressFinalize(this);

    foreach (Layer l in layers)
      l.Dispose();

    BonusPicture.Dispose();
    EnvelopePicture.Dispose();
    PlayerPicture.Dispose();
    ScorePicture.Dispose();

    titleScreen.Dispose();
    pauseScreen.Dispose();
    tryAgainScreen.Dispose();

    moveSound.Dispose();
    infectSound.Dispose();
    getBonusSound.Dispose();
    useBonusSound.Dispose();
    invalidMoveSound.Dispose();
    lostLifeSound.Dispose();
    extraLifeSound.Dispose();
  }

  // Returns the current state
  public static GameStates State
  {
    get { return state; }
    set
    {
      if (value == GameStates.NotStarted)
      {
        if (Score > HighScore) HighScore = Score;
      }
      else if (value == GameStates.TryAgain)
      {
        numLives--;

        // Check for game over
        if (numLives < 0)
        {
          // Call the property setter again
          State = GameStates.NotStarted;
          return;
        }
      }
      else if (value == GameStates.Running)
      {
        if (state != GameStates.Paused)
        {
          // Re-initialize the layers,
          // envelopes, and player
          foreach (Layer l in layers)
            l.Reset();

          foreach (Envelope m in EnvelopeGrid)
            m.Reset();

          Player.Reset();
        }

        if (state == GameStates.NotStarted)
        {
          score = 0;
          numLives = 3;
          numBonuses = 1;
          ScorePicture.SetPosition(Screen.Width,
            Screen.Height - DigitPadding - DigitHeight, false);
        }
      }
      state = value;
    }
  }

  // Returns the number of lives remaining
  public static int NumLives
  {
    get { return numLives; }
  }

  // Returns the number of bonuses remaining
  public static int NumBonuses
  {
    get { return numBonuses; }
  }

  // Returns the current score
  public static int Score
  {
    get { return score; }
  }

  // Gets/sets the highest score achieved on the machine
  public static int HighScore
  {
    get
    {
      if (highScore >= 0)
      {
        // We don't need to check the registry since
        // we must have obtained the value previously
        return highScore;
      }
      else
      {
        // Get the value from the registry.  If the value doesn't
        // already exist, set it to zero.
        RegistryKey key = null;
        object registryValue = null;

        try
        {
          key = Registry.LocalMachine.OpenSubKey(
            "SOFTWARE\\.NET and COM", true);

          if (key == null)
            key = Microsoft.Win32.Registry.LocalMachine.CreateSubKey(
              "SOFTWARE\\.NET and COM");

          registryValue = key.GetValue("HighScore");
          if (registryValue == null)
          {
            key.SetValue("HighScore", 0);
            registryValue = 0;
            highScore = 0;
          }
        }
        catch { highScore = 0;  return 0; }
        finally { if (key != null) key.Close(); }

        return (int)registryValue;
      }
    }
    set
    {
      // Save the high score
      highScore = value;

      // Place the high score in the registry.
      RegistryKey key = null;

      try
      {
        key = Registry.LocalMachine.OpenSubKey(
          "SOFTWARE\\.NET and COM", true);

        if (key == null)
          key = Microsoft.Win32.Registry.LocalMachine.CreateSubKey(
            "SOFTWARE\\.NET and COM");

        key.SetValue("HighScore", highScore);
      }
      catch { }
      finally { if (key != null) key.Close(); }
    }
  }

  // Moves the player and determines what to do afterwards
  public static void MovePlayer(Moves move)
  {
    if (!Player.IsDying && Player.Move(move))
    {
      // The player has moved, so check for the presence
      // of a visible envelope underneath.

      Envelope currentEnvelope = Player.EnvelopeUnderneath;

      if (currentEnvelope.IsVisible)
      {
        if (currentEnvelope.HasAttachment)
        {
          // Remove the attachment from the envelope and add it
          // to the player's list of bonuses

          currentEnvelope.HasAttachment = false;
          currentEnvelope.IsInfected = true;
          numBonuses++;

          // On the 11th bonus, remove all of them and add
          // an extra life
          if (numBonuses == 11)
          {
            numBonuses = 0;
            numLives++;
            score += 1000;
            extraLifeSound.Play();
          }
          else
          {
            score += 100;
            getBonusSound.Play();
          }
        }
        else if (!currentEnvelope.IsInfected)
        {
          currentEnvelope.IsInfected = true;
          score++;
          infectSound.Play();
        }
        else
        {
          // No points are given for stepping on an
          // already-infected envelope
          moveSound.Play();
        }
      }
      else
      {
        // If the envelope is invisible, the player is still alive
        // if there are any bonuses stored
        if (numBonuses > 0)
        {
          // Make a visible envelope right under the player
          currentEnvelope.MakeVisibleAndInfected();
          numBonuses--;
          useBonusSound.Play();
        }
        else
        {
          // The player dies
          KillPlayer();
        }
      }
    }
    else
    {
      invalidMoveSound.Play();
    }
  }

  // Check to see if the player died in the scenarios in which a key press
  // doesn't cause it: going off the left edge of the screen, and staying
  // on an envelope that disappears.
  public static void CheckForDeath()
  {
    if ((Player.LocationX + PlayerPicture.Width < 0) ||
      !Player.EnvelopeUnderneath.IsVisible)
    {
      KillPlayer();
    }
  }

  // Kill the player
  private static void KillPlayer()
  {
    lostLifeSound.Play();
    Player.Kill();
  }

  // Draw the contents of the current frame
  public static void Render()
  {
    if (state == GameStates.Running)
    {
      foreach (Layer l in layers)
        l.Render();

      Util.RenderComplete();
    }
    else if (state == GameStates.TryAgain)
    {
      tryAgainScreen.Render();
    }
    else if (state == GameStates.NotStarted)
    {
      titleScreen.Render();
    }
    else if (state == GameStates.Paused)
    {
      pauseScreen.Render();
    }
  }
}