using System;

public class Layer : IDisposable
{
  // Number of frames between each movement of one pixel
  public readonly int Frequency;  
  // Number of pixels that have been "scrolled" off the screen.
  // Resets to zero once an entire screen width has been scrolled.
  protected int pixelsOffscreen;  
  // The background image
  private Picture background;     

  // Creates the layer with the background picture
  // and initializes its characteristics
  public Layer(string bitmapFile, bool hasTransparentColor, int frequency)
  {
    background = new Picture(bitmapFile, hasTransparentColor);
    this.Frequency = frequency;
    pixelsOffscreen = 0;
  }

  // Disposes the picture
  ~Layer()
  {
    background.Dispose();
  }

  // Disposes the picture
  public void Dispose()
  {
    // Since we're cleaning up now, suppress finalization
    background.Dispose();
    GC.SuppressFinalize(this);
  }

  // Re-initializes the layer to the state when constructed
  public virtual void Reset()
  {
    // In the base implementation, simply reset the # of pixels scrolled
    pixelsOffscreen = 0;
  }

  // Draw the contents of the layer.
  // This must be followed by a call to Util.RenderComplete().
  public virtual void Render()
  {
    if (Frequency == 0)
    {
      // Always draw the background in the top left corner
      background.DrawAt(0, 0);
    }
    else
    {
      // Check if it's time to update the # of pixels scrolled
      if (Util.IsTimeToMove(Frequency))
        pixelsOffscreen = (pixelsOffscreen + 1) % Screen.Width;

      // Draw the background picture consecutively until the screen is
      // filled horizontally
      for (int i = -pixelsOffscreen; i < Screen.Width; i += background.Width)
      {
        // Draw the background at the current offset
        background.DrawAt(i, 0);
      }
    }
  }
}