using DirectX;
using System;
using System.Runtime.InteropServices;

public class Picture : IDisposable
{
  protected DirectDrawSurface7 dxSurface;  // The DirectX surface
  protected RECT size;                     // Contains the dimensions
  protected bool hasTransparentColor;      // True if the input bitmap
                                           // uses a transparent color

  // Creates the picture from the bitmap
  public Picture(string bitmapFile, bool hasTransparentColor)
  {
    this.hasTransparentColor = hasTransparentColor;
    size = new RECT();
    dxSurface = Util.CreateSurfaceFromFile(
      bitmapFile, hasTransparentColor, ref size);
  }

  // Disposes the picture
  ~Picture()
  {
    Marshal.ReleaseComObject(dxSurface);
  }

  // Disposes the picture
  public void Dispose()
  {
    // Since we're cleaning up now, suppress finalization
    Marshal.ReleaseComObject(dxSurface);
    GC.SuppressFinalize(this);
  }

  // Returns the width of the picture
  public virtual int Width
  {
    get { return size.Right - size.Left; }
  }

  // Returns the height the picture
  public virtual int Height
  {
    get { return size.Bottom - size.Top; }
  }

  // Returns the input rectangle, but "clipped" if part of it
  // is located outside the range of the screen.
  // The coordinates are changed to zero if they are negative.
  protected RECT ClippedRectangle(ref int x, ref int y, RECT size)
  {
    RECT r = size;

    // Check if the picture would be off the left edge of the screen
    if (x < 0)
    {
      r.Left -= x;
      x = 0;
    }
    // Check if the picture would be off the right edge of the screen
    else if (x + (r.Right - r.Left) >= Screen.Width)
    {
      r.Right = r.Left + Screen.Width - x;
    }

    // Check if the picture would be off the top edge of the screen
    if (y < 0)
    {
      r.Top -= y;
      y = 0;
    }
    // Check if the picture would be off the bottom edge of the screen
    else if (y + (r.Bottom - r.Top) >= Screen.Height)
    {
      r.Bottom = r.Top + Screen.Height - y;
    }

    return r;
  }

  // Draw the picture with the top left corner at the
  // (x,y) coordinate.  The picture is clipped if part
  // of the picture is off the screen.
  public virtual void DrawAt(int x, int y)
  {
    RECT r = ClippedRectangle(ref x, ref y, size);
    Util.DrawSurface(dxSurface, x, y, r, hasTransparentColor);
  }

  // Draw the picture with the top left corner at the
  // (x,y) coordinate, resized to the width and height given.
  // The picture is clipped if part of the picture is off
  // the screen.
  public virtual void ResizeAndDrawAt(int x, int y, int width, int height)
  {
    RECT r = ClippedRectangle(ref x, ref y, size);

    RECT dest = new RECT();
    dest.Left = x;
    dest.Top = y;
    dest.Right = x + width;
    dest.Bottom = y + height;

    Util.ResizeAndDrawSurface(dxSurface, r, dest, hasTransparentColor);
  }
}