using DirectX;
using System;
using System.Runtime.InteropServices;

public class Util
{
  // A counter that is updated with every frame
  public static int Counter;  
  // A pseudo-random number
  public static Random RandomNumber;  

  // Main DirectX objects
  private static IDirectX7 directX;
  private static DirectDraw7 directDraw;
  private static DirectSound directSound;

  // Two surfaces used for double-buffered animation
  private static DirectDrawSurface7 frontBuffer;
  private static DirectDrawSurface7 backBuffer;

  // This must be called before using any of the other Util methods
  public static void Initialize(IntPtr hWnd)
  {
    // Create the main DirectX objects
    directX = (IDirectX7)new DirectX7();
    directDraw = directX.DirectDrawCreate("");
    directSound = directX.DirectSoundCreate("");

    // Set our cooperative level for drawing and sound
    directDraw.SetCooperativeLevel(hWnd.ToInt32(),
      CONST_DDSCLFLAGS.DDSCL_EXCLUSIVE |
      CONST_DDSCLFLAGS.DDSCL_FULLSCREEN);
    directSound.SetCooperativeLevel(hWnd.ToInt32(), 
      CONST_DSSCLFLAGS.DSSCL_NORMAL);

    // Set the display mode
    directDraw.SetDisplayMode(Screen.Width, Screen.Height, 16, 0, 0);
    // Create the primary surface with one back buffer
    DDSURFACEDESC2 description = new DDSURFACEDESC2();
    DDSCAPS2 capabilities = new DDSCAPS2();

    description.lSize = Marshal.SizeOf(description);
    description.lFlags = CONST_DDSURFACEDESCFLAGS.DDSD_CAPS |
                         CONST_DDSURFACEDESCFLAGS.DDSD_BACKBUFFERCOUNT;
    description.ddsCaps.lCaps = 
      CONST_DDSURFACECAPSFLAGS.DDSCAPS_PRIMARYSURFACE |
      CONST_DDSURFACECAPSFLAGS.DDSCAPS_FLIP |
      CONST_DDSURFACECAPSFLAGS.DDSCAPS_COMPLEX;
    description.lBackBufferCount = 1;

    frontBuffer = directDraw.CreateSurface(ref description);
    // Get the single back buffer.
    capabilities.lCaps = CONST_DDSURFACECAPSFLAGS.DDSCAPS_BACKBUFFER;
    capabilities.lCaps2 = 0;
    capabilities.lCaps3 = 0;
    capabilities.lCaps4 = 0;
    backBuffer = frontBuffer.GetAttachedSurface(ref capabilities);

    // Initialize the pseudo-random number generator.
    RandomNumber = new Random();
  }

  // Creates a surface from the input bitmap
  public static DirectDrawSurface7 CreateSurfaceFromFile(
    string bitmapFile, bool hasTransparentColor, ref RECT size)
  {
    DirectDrawSurface7 dxSurface = null;

    try
    {
      // Create the DirectDraw surface
      DDSURFACEDESC2 description = new DDSURFACEDESC2();
      description.lSize = Marshal.SizeOf(description);
      description.lFlags = CONST_DDSURFACEDESCFLAGS.DDSD_CAPS |
        CONST_DDSURFACEDESCFLAGS.DDSD_WIDTH | 
        CONST_DDSURFACEDESCFLAGS.DDSD_HEIGHT;
      description.ddsCaps.lCaps = 
        CONST_DDSURFACECAPSFLAGS.DDSCAPS_OFFSCREENPLAIN;
      dxSurface = directDraw.CreateSurfaceFromFile(
        bitmapFile, ref description);

      // Save the picture's dimensions
      size.Left = 0;
      size.Top = 0;
      size.Right = description.lWidth;
      size.Bottom = description.lHeight;

      if (hasTransparentColor)
      {
        // Get the pixel format
        DDPIXELFORMAT format = new DDPIXELFORMAT();
        format.lSize=Marshal.SizeOf(format);
        dxSurface.GetPixelFormat(ref format);

        // Set the color key to Red
        int KeyColor = format.lRBitMask;
        DDCOLORKEY key = new DDCOLORKEY();
        key.low = KeyColor;
        key.high = KeyColor;
        dxSurface.SetColorKey(CONST_DDCKEYFLAGS.DDCKEY_SRCBLT, ref key);
      }
    }
    catch (Exception e)
    {
      throw new ApplicationException("Failed to create surface for " + 
        bitmapFile + ".", e);
    }
    return dxSurface;
  }

  // Returns true if it's time to move, based on the
  // global counter and input frequency.
  public static bool IsTimeToMove(int frequency)
  {
    if (Counter % frequency == 0)
      return true;
    else
      return false;
  }

  // After drawing to the back buffer with DrawSurface or
  // ResizeAndDrawSurface, this must be called to show the
  // result on the front buffer.
  public static void RenderComplete()
  {
    try
    {
      frontBuffer.Flip(null, CONST_DDFLIPFLAGS.DDFLIP_WAIT);
    }
    catch
    {
        try { directDraw.RestoreAllSurfaces(); }
        catch { // Attempt to continue
        }
    }
  }

  // Draw the portion of the surface specified with r at location (x,y).
  // This must eventually be followed by a call to RenderComplete.
  public static void DrawSurface(DirectDrawSurface7 surface, int x,
    int y, RECT r, bool hasTransparentColor)
  {
    CONST_DDBLTFASTFLAGS transparentFlagFast;

    if (hasTransparentColor)
      transparentFlagFast = CONST_DDBLTFASTFLAGS.DDBLTFAST_SRCCOLORKEY;
    else
      transparentFlagFast = 0;

    try
    {
      backBuffer.BltFast(x, y, surface, ref r,
        CONST_DDBLTFASTFLAGS.DDBLTFAST_WAIT | transparentFlagFast);
    }
    catch
    {
        try { directDraw.RestoreAllSurfaces(); }
        catch { // Attempt to continue
        }
    }
  }

  // Draw the portion of the surface specified with the RECT dest using
  // the location and size specified in the RECT source.
  // This must eventually be followed by a call to RenderComplete.
  public static void ResizeAndDrawSurface(DirectDrawSurface7 surface,
    RECT source, RECT dest, bool hasTransparentColor)
  {
    CONST_DDBLTFLAGS transparentFlag;

    if (hasTransparentColor)
      transparentFlag = CONST_DDBLTFLAGS.DDBLT_KEYSRC;
    else
      transparentFlag = 0;

    try
    {
      backBuffer.Blt(ref dest, surface, ref source, 
        CONST_DDBLTFLAGS.DDBLT_WAIT | transparentFlag);
    }
    catch
    {
        try { directDraw.RestoreAllSurfaces(); }
        catch { // Attempt to continue
        }
    }
  }

  // Create a sound buffer from the input .WAV file.
  public static DirectSoundBuffer 
    CreateSoundBufferFromFile(string waveFile)
  {
    try
    {
      WAVEFORMATEX format = new WAVEFORMATEX();

      DSBUFFERDESC description = new DSBUFFERDESC();
      description.lSize = Marshal.SizeOf(description);
      description.lFlags = CONST_DSBCAPSFLAGS.DSBCAPS_CTRLFREQUENCY |
        CONST_DSBCAPSFLAGS.DSBCAPS_CTRLPAN |
        CONST_DSBCAPSFLAGS.DSBCAPS_CTRLVOLUME | 
        CONST_DSBCAPSFLAGS.DSBCAPS_STATIC;
      return directSound.CreateSoundBufferFromFile(waveFile,
        ref description, out format);
    }
    catch (Exception e)
    {
      throw new ApplicationException("Failed to create buffer for " 
        + waveFile + ".", e);
    }
  }

  // Play a sound buffer
  public static void PlayBuffer(DirectSoundBuffer buffer)
  {
    CONST_DSBSTATUSFLAGS dwStatus = buffer.GetStatus();

    // Ensure that the buffer isn't lost
    if ((dwStatus & CONST_DSBSTATUSFLAGS.DSBSTATUS_BUFFERLOST) != 0)
      buffer.restore();

    buffer.Play(CONST_DSBPLAYFLAGS.DSBPLAY_DEFAULT);
  }
}