// To compile:
// csc /t:library DancingCatVisualization.cs /r:System.Drawing.dll /r:BookExamples.WindowsMediaPlayer.dll /res:DancingCatVisualization.resources

using System;
using System.Resources;
using System.Drawing;
using System.Drawing.Drawing2D;
using BookExamples.WindowsMediaPlayer;

public class DancingCatVisualization : Visualization, IDisposable
{
  private Bitmap [,] catPictures;

  public DancingCatVisualization()
  {
    ResourceManager resources = 
      new ResourceManager(typeof(DancingCatVisualization));
    catPictures = new Bitmap[2,2];
    catPictures[0,0] = (Bitmap)resources.GetObject("Left1");
    catPictures[0,1] = (Bitmap)resources.GetObject("Left2");
    catPictures[1,0] = (Bitmap)resources.GetObject("Right1");
    catPictures[1,1] = (Bitmap)resources.GetObject("Right2");

    foreach (Bitmap b in catPictures)
      b.MakeTransparent(Color.Red);
  }

  ~DancingCatVisualization()
  {
    foreach (Bitmap b in catPictures)
      b.Dispose();
  }

  public void Dispose() 
  {
    // Since we're cleaning up now, suppress finalization.
    GC.SuppressFinalize(this);

    foreach (Bitmap b in catPictures)
      b.Dispose();
  }

  public override string Title
  {
    get { return "Dancing Cat"; }
  }

  public override string Description
  {
    get { return "A visualization with a cat that dances to the music."; }
  }

  public override void Render(TimedLevel levels, Graphics g, Rectangle r)
  {
    Bitmap cat;

    switch (levels.State)
    {
      case AudioState.Stopped:
      {
        using (Brush b = new LinearGradientBrush(r,
          Color.White, Color.DarkGray, LinearGradientMode.Vertical))
        {
          // Fill background with a gradient
          g.FillRectangle(b, r);
        }

        cat = catPictures[0,0];
        g.DrawImage(cat, (r.Width - cat.Width) / 2, 
          (r.Height - cat.Height) / 2);
      }
      break;

      case AudioState.Paused:
      {
        using (Brush b = new LinearGradientBrush(r,
          Color.Olive, Color.LightYellow, LinearGradientMode.Vertical))
        {
          // Fill background with a gradient
          g.FillRectangle(b, r);
        }

        cat = catPictures[0,0];
        g.DrawImage(cat, 
          (r.Width - cat.Width) / 2, (r.Height - cat.Height) / 2);
      }
      break;

      case AudioState.Playing:
      {
        using (LinearGradientBrush b = new LinearGradientBrush(r,
          Color.LightYellow, Color.Olive, LinearGradientMode.Vertical))
        {
          // Fill background with a gradient
          g.FillRectangle(b, r);
        }

        // Alternate orientation and stance at regular intervals
        int orientation = (int)((levels.TimeStamp / 10000000) % 2);
        int stance = (int)((levels.TimeStamp / 1000000) % 2);

        cat = catPictures[orientation,stance];
        g.DrawImage(cat, 
          (r.Width - cat.Width) / 2 
          + (levels.GetFrequency(AudioChannel.Left, 0) - 128) / 2, 
          (r.Height - cat.Height) / 2 
          + (levels.GetWaveform(AudioChannel.Left, 0) - 128) / 2);
      }
      break;
    }
  }
}