// This is to be compiled with a few other source files as follows:
//
// csc /t:library /out:BookExamples.WindowsMediaPlayer.dll IWMPEffects.cs SupportingTypes.cs Visualization.cs PresetEnumAttribute.cs /r:System.Drawing.dll /r:System.Windows.Forms.dll /r:System.dll

using System;
using System.Runtime.InteropServices;
using System.Drawing;
using System.Windows.Forms;
using System.Reflection;
using Microsoft.Win32;

namespace BookExamples.WindowsMediaPlayer
{
  public abstract class Visualization : IWMPEffects
  {
    private int channelCount;
    private int sampleRate;
    private string audioTitle;
    private int currentPreset;

    // Used for double buffering in full-screen mode
    Graphics offScreenGraphics;
    Graphics onScreenGraphics;
    Bitmap offScreenBitmap;
    Rectangle fullScreenRect;

    [DllImport("user32.dll", ExactSpelling=true)]
    private static extern IntPtr GetForegroundWindow();

    [DllImport("user32.dll", ExactSpelling=true)]
    private static extern bool GetWindowRect(IntPtr hwnd, out RECT r);

    // Custom registration
    [ComRegisterFunction]
    private static void ComRegisterFunction(Type t)
    {
      Visualization v = (Visualization)Activator.CreateInstance(t);
      RegistryKey key = Registry.LocalMachine;
      key = key.CreateSubKey(
        "Software\\Microsoft\\MediaPlayer\\Objects\\Effects\\"
        + v.Title + "\\Properties");
      key.SetValue("classid", t.GUID.ToString("B"));
      key.SetValue("name", v.Title);
      key.SetValue("description", v.Description);
      key.Close();
    }

    // Custom unregistration
    [ComUnregisterFunction]
    private static void ComUnregisterFunction(Type t)
    {
      Visualization v = (Visualization)Activator.CreateInstance(t);
      RegistryKey key = Registry.LocalMachine;
      key.DeleteSubKey(
        "Software\\Microsoft\\MediaPlayer\\Objects\\Effects\\" 
        + v.Title + "\\Properties");
      key.DeleteSubKey(
        "Software\\Microsoft\\MediaPlayer\\Objects\\Effects\\" 
        + v.Title);
      key.Close();
    }

    // The current number of audio channels
    public int CurrentChannelCount
    {
      get { return channelCount; }
    }

    // The current sample rate of the audio
    public int CurrentSampleRate
    {
      get { return sampleRate; }
    }

    // The title of the currently playing audio
    public string CurrentAudioTitle
    {
      get { return audioTitle; }
    }

    // The number of the current preset
    public virtual int CurrentPreset
    {
      get { return currentPreset; }
      set { currentPreset = value; }
    }
 
    // A description of the visualization
    public virtual string Description
    {
      get { return "A .NET visualization."; }
    }

    // The title of visualization
    public abstract string Title { get; }

    // Event raised when the current media changes.
    // Information about the media can be accessed through
    // the CurrentChannelCount, CurrentSampleRate, and 
    // CurrentAudioTitle properties.
    public event EventHandler MediaChange;

    // Event raised when the visualization enters full-screen mode
    public event EventHandler EnterFullScreen;

    // Event raised when the visualization exits full-screen mode
    public event EventHandler ExitFullScreen;

    // The original Render method
    void IWMPEffects.Render(
      ref TimedLevel levels, IntPtr hdc, ref RECT r)
    {
      using (Graphics g = Graphics.FromHdc(hdc))
      {
        Render(levels, g, Rectangle.FromLTRB(
          r.Left, r.Top, r.Right, r.Bottom));
      }
    }

    // The public Render method
    public abstract void Render(
      TimedLevel levels, Graphics g, Rectangle r);

    // The original MediaInfo method
    void IWMPEffects.MediaInfo(
      int channelCount, int sampleRate, string title)
    {
      this.channelCount = channelCount;
      this.sampleRate = sampleRate;
      this.audioTitle = title;

      OnMediaChange(new EventArgs());
    }

    // Raises the MediaChange event
    protected virtual void OnMediaChange(EventArgs e)
    {
      if (MediaChange != null) MediaChange(this, e);
    }

    // Returns CanGoFullScreen or CanGoFullScreen | HasPropertyPage
    // if the current instance overrides DisplayPropertyPage
    public virtual VisualizationCapabilities Capabilities
    {
      get 
      {
        if (this.GetType().GetMethod("DisplayPropertyPage",
          BindingFlags.Public | 
          BindingFlags.Instance |
          BindingFlags.DeclaredOnly) == null)
        {
          return VisualizationCapabilities.CanGoFullScreen; 
        }
        else
        {
          return VisualizationCapabilities.CanGoFullScreen |
                 VisualizationCapabilities.HasPropertyPage; 
        }
      }
    }

    // Returns the title of the visualization unless the current
    // instance uses PresetEnumAttribute to define its presets
    public virtual string GetPresetTitle(int preset)
    {
      // Check for the PresetEnumAttribute
      PresetEnumAttribute [] attributes = (PresetEnumAttribute []) 
        this.GetType().GetCustomAttributes(
        typeof(PresetEnumAttribute), true);

      if (attributes.Length == 0)
      {
        // By default, return the title as the name of the preset.
        return Title;
      }
      else
      {
        if (!Enum.IsDefined(attributes[0].EnumType, preset))
          throw new ArgumentOutOfRangeException("preset");

        return Enum.Format(attributes[0].EnumType, preset, "G");
      }
    }

    // Returns 1 unless the current instance uses
    // PresetEnumAttribute to define its presets
    public virtual int PresetCount
    {
      get 
      { 
        // Check for the PresetEnumAttribute
        PresetEnumAttribute [] attributes = (PresetEnumAttribute [])
          this.GetType().GetCustomAttributes(
          typeof(PresetEnumAttribute), true);
	
        if (attributes.Length == 0)
        {
          // By default, return only one preset.
          return 1;
        }
        else
        {
          return Enum.GetValues(attributes[0].EnumType).Length; 
        }
      }
    }

    // The original DisplayPropertyPage method
    void IWMPEffects.DisplayPropertyPage(IntPtr hwndOwner)
    {
      OwnerWindow owner = new OwnerWindow();
      owner.Handle = hwndOwner;
      DisplayPropertyPage(owner);
    }

    // The public DisplayPropertyPage method
    public virtual void DisplayPropertyPage(IWin32Window owner)
    {
        throw new NotImplementedException();
    }

    // Called when switching between regular and full screen modes
    void IWMPEffects.GoFullScreen(bool startFullScreen)
    {
      if (startFullScreen)
      {
        IntPtr handle = GetForegroundWindow();
        RECT r;
        GetWindowRect(handle, out r);

        fullScreenRect = Rectangle.FromLTRB(
          r.Left, r.Top, r.Right, r.Bottom);

        // Create an off-screen bitmap and a Graphics object for it
        offScreenBitmap = new Bitmap(r.Right-r.Left, r.Bottom-r.Top);
        offScreenGraphics = Graphics.FromImage(offScreenBitmap);

        // Construct a Graphics object from the HWND
        onScreenGraphics = Graphics.FromHwnd(handle);

        OnEnterFullScreen(new EventArgs());
      }
      else
      {
        // Dispose the off-screen bitmap and Graphics objects
        offScreenBitmap.Dispose();
        offScreenGraphics.Dispose();
        onScreenGraphics.Dispose();

        OnExitFullScreen(new EventArgs());
      }
    }

    // Raises the EnterFullScreen event
    protected virtual void OnEnterFullScreen(EventArgs e)
    {
      if (EnterFullScreen != null) EnterFullScreen(this, e);
    }

    // Raises the ExitFullScreen event
    protected virtual void OnExitFullScreen(EventArgs e)
    {
      if (ExitFullScreen != null) ExitFullScreen(this, e);
    }

    // Called during full-screen rendering
    void IWMPEffects.RenderFullScreen(ref TimedLevel levels)
    {
      Render(levels, offScreenGraphics, fullScreenRect);

      // Draw the entire off-screen bitmap on the screen
      onScreenGraphics.DrawImage(offScreenBitmap, 0, 0);
    }
  }

  // Used by Visualization's DisplayPropertyPage to construct
  // the managed view of a window with a certain HWND.
  class OwnerWindow : IWin32Window
  {
    private IntPtr handle;

    public IntPtr Handle
    {
      get { return handle; }
      set { handle = value; }
    }
  }
}