using System;

/// <summary>
/// Type library compatibility checker.
/// </summary>
class TlbCompat
{
	static string oldTlbName = null;
	static string newTlbName = null;
	static TypeLibComparer comparer = null;

	/// <summary>
	/// The main entry point for the application.
	/// </summary>
	private static int Main(string[] args)
	{
		PrintHeader();

		if (args.Length <= 1)
		{
			PrintUsage();
			return (int)ReturnCodes.Help;
		}
		else
		{
			oldTlbName = args[0];
			newTlbName = args[1];
		}

		try
		{
			comparer = new TypeLibComparer(oldTlbName, newTlbName);

			if (args.Length > 2 && (args[2] == "/errorsonly" || args[2] == "-errorsonly"))
				comparer.ErrorsOnly = true;

			comparer.CheckCompatibility();
			comparer.PrintSummary();
		}
		catch (ApplicationException e)
		{
			PrintError(e.Message);
			return (int)ReturnCodes.FatalError;
		}

		// Return a code representing the most severe thing that happens
		if (comparer.NumErrors > 0)
            return (int)ReturnCodes.CompatErrors;
		else if (comparer.NumWarnings > 0)
			return (int)ReturnCodes.CompatWarnings;
		else if (comparer.NumOtherDiffs > 0)
			return (int)ReturnCodes.CompatOtherDiffs;
		else
			return (int)ReturnCodes.Success;
	}

	/// <summary>
	/// Print a description of using this utility.
	/// </summary>
	private static void PrintUsage()
	{
		Console.WriteLine("Syntax: TlbCompat <OldTypeLib> <NewTypeLib> [Options]");
		Console.WriteLine("");
		Console.WriteLine("This tool ensures that the type library specified by NewTypeLib");
		Console.WriteLine("is compatible with the type library specified by OldTypeLib.");
		Console.WriteLine("It is designed to be used with type libraries exported from .NET assemblies.");
		Console.WriteLine("");
		Console.WriteLine("Options:");
		Console.WriteLine("    /errorsonly     Don't display individual warnings or other differences");
		Console.WriteLine("");
		Console.WriteLine("It produces three kinds of messages:");
		Console.WriteLine("");
		Console.WriteLine("  1. Errors - changes that break compatibility.");
		Console.WriteLine("  2. Warnings - changes that could break compatibility, based on installation.");
		Console.WriteLine("                These are only given for GUID changes of non-interfaces.");
		Console.WriteLine("  3. Other Diffs - minor changes should not break compatibility.");
		Console.WriteLine("");
		Console.WriteLine("It determines that NewTypeLib is compatible with OldTypeLib if:");
		Console.WriteLine("");
		Console.WriteLine("  1. NewTypeLib has at least all the types that OldTypeLib has, with the");
		Console.WriteLine("     same names and mostly the same flags.  A change in GUID for a");
		Console.WriteLine("     non-interface only produces a warning, because it can be compatible if");
		Console.WriteLine("     the old types are still registered under their old GUIDs, and policy");
		Console.WriteLine("     redirects requests for the old types to the new types.");
		Console.WriteLine("  2. All interfaces in NewTypeLib have the same IIDs as they do in OldTypeLib.");
		Console.WriteLine("  3. All coclasses from OldTypeLib list at least the same implemented");
		Console.WriteLine("     interfaces and source interfaces in NewTypeLib, and have the same");
		Console.WriteLine("     default interfaces.");
		Console.WriteLine("  4. Interfaces from OldTypeLib look identical in NewTypeLib (although");
		Console.WriteLine("     parameter names are allowed to change case), ignoring help-related info");
		Console.WriteLine("     such as helpstrings and ignoring IDL custom attributes.");
		Console.WriteLine("     (An error also is reported if dispinterface methods change ordering,");
		Console.WriteLine("      which is more strict than necessary.)");
	}

	/// <summary>
	/// Print the logo.
	/// </summary>
	private static void PrintHeader()
	{
		Console.WriteLine("Type Library Compatibility Verifier 1.0.0.0");
		Console.WriteLine("");
	}

	/// <summary>
	/// Print a fatal error.
	/// </summary>
	/// <param name="message">The message to print</param>
	private static void PrintError(string message)
	{
		Console.WriteLine("FATAL ERROR: " + message);
	}
}

internal enum ReturnCodes
{
	Help = 1,
	Success = 2,
	FatalError = 3,
	CompatErrors = 4,
	CompatWarnings = 5,
	CompatOtherDiffs = 6
}